class MissingActiveToken(Exception):
    def __init__(self, token=None, message="Missing Active Token") -> None:
        self.token = token
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.token} -> {self.message}"



import argparse
import json

from tabulate import tabulate

from pipeline.api import PipelineCloud
from pipeline.schemas.file import FileGet
from pipeline.schemas.pagination import Paginated
from pipeline.schemas.run import RunGet, RunState
from pipeline.util import hex_to_python_object


def runs(args: argparse.Namespace) -> int:
    sub_command = getattr(args, "sub-command", None)

    remote_service = PipelineCloud(verbose=False)
    remote_service.authenticate()

    if sub_command in ["list", "ls"]:
        raw_result = remote_service.get_runs()

        schema = Paginated[RunGet].parse_obj(raw_result)

        runs = schema.data

        terminal_run_states = [
            RunState.FAILED,
            RunState.COMPLETE,
        ]

        run_data = [
            [
                _run.id,
                _run.created_at.strftime("%d-%m-%Y %H:%M:%S"),
                "executing",
                _run.runnable.id,
            ]
            for _run in runs
            if _run.run_state not in terminal_run_states
        ]
        table = tabulate(
            run_data,
            headers=[
                "ID",
                "Created at",
                "State",
                "Pipeline",
            ],
            tablefmt="outline",
        )
        print(table)
        return 0
    elif sub_command == "get":
        run_id = args.run_id

        result = remote_service._get(f"/v2/runs/{run_id}")
        if args.result:
            result = RunGet.parse_obj(result)
            if result.result_preview is not None:
                print(json.dumps(result.result_preview))
            else:
                file_schema_raw = remote_service._get(
                    f"/v2/files/{result.result.id}?return_data=true"
                )

                file_schema = FileGet.parse_obj(file_schema_raw)
                raw_result = hex_to_python_object(file_schema.data)
                print(json.dumps(raw_result))

            return 0
        else:
            print(json.dumps(result))
            return 0



from typing import Optional

from .base import BaseModel


class MilestonesRegisterGet(BaseModel):
    """View of a users' milestones register"""

    #: Has the user ever deployed a pipeline?
    pipeline_deployed: bool
    #: Has the user ever executed a pipeline run?
    run_executed: bool
    #: Has the user ever visited dashboard modelhub?
    modelhub_viewed: bool


class MilestonesRegisterPatch(BaseModel):
    """Patch a users' milestones register"""

    # `pipeline_deployed` & `run_executed` update managed internally

    modelhub_viewed: Optional[bool]



import datetime
from enum import Enum
from typing import Any, List, Optional, Union

from pydantic import root_validator, validator

from pipeline.schemas.compute_requirements import ComputeRequirements, ComputeType
from pipeline.schemas.file import FileGet
from pipeline.schemas.function import FunctionGet, FunctionGetDetailed
from pipeline.schemas.pipeline import PipelineGet, PipelineGetDetailed

from .base import BaseModel
from .data import DataGet
from .runnable import RunnableIOGet
from .tag import TagGet
from .token import TokenGet


class RunState(Enum):
    RECEIVED = "received"
    ALLOCATING_CLUSTER = "allocating_cluster"
    AWAITING_RESOURCE_ALLOCATION = "awaiting_resource_allocation"
    ALLOCATING_RESOURCE = "allocating_resource"
    LOADING_DATA = "loading_data"
    LOADING_FUNCTION = "loading_function"
    LOADING_PIPELINE = "loading_pipeline"
    RUNNING = "running"
    COMPLETE = "complete"
    FAILED = "failed"


class RunErrorType(Enum):
    """The type of error that occurred (if any)"""

    MAX_RETRIES = "max_retries"
    PIPELINE_FAULT = "pipeline_fault"
    UNSATISFIABLE = "unsatisfiable"


class RunErrorInfo(BaseModel):
    """More info about the error if it was a pipeline_fault"""

    exception: str
    traceback: str


class RunCreate(BaseModel):
    pipeline_id: Optional[str]
    function_id: Optional[str]
    data: Optional[Any]
    data_id: Optional[str]
    blocking: Optional[bool] = False
    compute_type: Optional[ComputeType]
    compute_requirements: Optional[ComputeRequirements]

    @root_validator
    def pipeline_data_val(cls, values):
        pipeline_id, function_id = values.get("pipeline_id"), values.get("function_id")

        pipeline_defined = pipeline_id is not None
        function_defined = function_id is not None

        if pipeline_defined == function_defined:
            raise ValueError("You must define either a pipeline_id OR function_id.")

        data_id, data = values.get("data_id"), values.get("data")

        data_defined = data is not None
        data_id_defined = data_id is not None

        if data_defined == data_id_defined:
            raise ValueError("You must define either a data_id OR data.")

        return values


class RunIOGet(RunnableIOGet):
    """Realised/given input/output data to a Runnable used in a Run."""

    value: str
    data_url: str


class RunGet(BaseModel):
    id: str
    created_at: datetime.datetime
    started_at: Optional[datetime.datetime]
    ended_at: Optional[datetime.datetime]
    run_state: RunState
    resource_type: Optional[str]
    compute_time_ms: Optional[int]
    runnable: Union[FunctionGet, PipelineGet]
    data: DataGet
    blocking: Optional[bool] = False
    result: Optional[FileGet]
    #: JSON-serialised runnable return value, if available
    result_preview: Optional[Union[list, dict]]
    error: Optional[RunErrorType]
    error_info: Optional[RunErrorInfo]
    compute_requirements: Optional[ComputeRequirements] = None

    class Config:
        allow_population_by_field_name = True

    @validator("compute_requirements", pre=True)
    def compute_requirements_default_if_empty(cls, v):
        """Return None if no compute_requirements rather than an empty
        ComputeRequirements object
        """
        if not v:
            return None
        return v


class RunGetDetailed(RunGet):
    runnable: Union[FunctionGetDetailed, PipelineGetDetailed]
    n_resources: int
    region: str
    tags: List[TagGet] = []
    inputs: List[RunIOGet] = []
    outputs: List[RunIOGet] = []
    #: The Token which was used to create this Run, if that Token has not been
    #: deleted in the meantime
    token: Optional[TokenGet]


class RunUpdate(BaseModel):
    result_id: Optional[str]
    run_state: Optional[RunState]
    compute_cluster_id: Optional[str]



"""Request/response schemas for user onboarding tasks."""
from typing import Optional

from pydantic import HttpUrl

from .base import BaseModel


class OnboardingTaskGet(BaseModel):
    """Response schema describing a single onboarding task."""

    #: Database identifier
    id: str
    #: Shorthand name; unique across tasks
    name: str
    #: Longform description
    description: Optional[str]
    #: Approximate time required to complete this task
    time_to_complete_min: int
    #: URL to further information
    url: Optional[HttpUrl]
    #: Completion status
    complete: bool


class OnboardingTaskPatch(BaseModel):
    """Patch schema for updating a single onboarding task."""

    #: Completion status
    complete: bool



from typing import Optional

from .base import BaseModel


class RequestGet(BaseModel):
    id: str
    token: Optional[str]
    status: str
    request_json: str
    result_json: str
    result_file_url: Optional[str]
    resource_url: str
    request_method: str
    time_requested: float
    request_duration: int
    ip_address: str



from __future__ import annotations

import hashlib
import io
import json
import os
import sys
import urllib.parse
import uuid
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Type, Union

import dill
import httpx
from pydantic import ValidationError
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from pipeline import configuration
from pipeline.exceptions.InvalidSchema import InvalidSchema
from pipeline.exceptions.MissingActiveToken import MissingActiveToken
from pipeline.objects.variable import PipelineFile
from pipeline.schemas.base import BaseModel
from pipeline.schemas.compute_requirements import ComputeRequirements
from pipeline.schemas.data import DataGet
from pipeline.schemas.file import FileCreate, FileFormat, FileGet
from pipeline.schemas.function import FunctionCreate, FunctionGet
from pipeline.schemas.model import ModelCreate, ModelGet
from pipeline.schemas.pipeline import (
    PipelineCreate,
    PipelineFileVariableGet,
    PipelineGet,
    PipelineVariableGet,
)
from pipeline.schemas.pipeline_file import (
    MultipartUploadMetadata,
    PipelineFileDirectUploadFinaliseCreate,
    PipelineFileDirectUploadInitCreate,
    PipelineFileDirectUploadInitGet,
    PipelineFileDirectUploadPartCreate,
    PipelineFileDirectUploadPartGet,
    PipelineFileGet,
)
from pipeline.schemas.run import RunCreate, RunGet
from pipeline.util import (
    CallbackBytesIO,
    generate_id,
    hex_to_python_object,
    python_object_to_hex,
    python_object_to_name,
)
from pipeline.util.logging import PIPELINE_FILE_STR, PIPELINE_STR

if TYPE_CHECKING:
    from pipeline.objects import Function, Graph, Model

FILE_CHUNK_SIZE = 200 * 1024 * 1024  # 200 MiB


class PipelineCloud:
    token: Optional[str]
    url: Optional[str]

    def __init__(
        self,
        *,
        url: str = None,
        token: str = None,
        timeout: float = 60.0,
        verbose: bool = True,
    ):
        self.url = url or os.getenv("PIPELINE_API_URL", configuration.DEFAULT_REMOTE)

        self.token = (
            token
            or os.getenv("PIPELINE_API_TOKEN")
            or configuration.remote_auth.get(self.url)
        )

        self.timeout = timeout
        self.verbose = verbose
        self.__valid_token__ = False
        if self.token is not None:
            self.authenticate()

    def authenticate(self, token: str = None) -> None:
        """
        Authenticate with the pipeline.ai API
            Parameters:
                token (str): API user token for authentication.
                    Pass it as an arg or set it as an ENV var.
            Returns:
                None
        """
        if self.verbose:
            print("Authenticating")

        _token = token or self.token
        if _token is None:
            raise MissingActiveToken(
                token="", message="Please pass a valid token or set it as an ENV var"
            )

        valid_token = self._validate_token(_token, self.url)

        if not valid_token:
            raise MissingActiveToken(token=_token)
        elif self.verbose:
            print("Succesfully authenticated with the Pipeline API (%s)" % self.url)

        self.token = _token
        self.__valid_token__ = True

    def raise_for_invalid_token(self):
        if not self.__valid_token__:
            raise MissingActiveToken(
                token=self.token,
                message=(
                    "Please set a valid token as an ENV var "
                    "and call authenticate(); "
                    "or pass a valid token as a parameter authenticate(token)"
                ),
            )

    @staticmethod
    def _validate_token(token: str, base_url: str) -> bool:
        url = urllib.parse.urljoin(base_url, "/v2/users/me")

        headers = {"Authorization": f"Bearer {token}"}
        response = httpx.get(url, headers=headers)
        if response.status_code == HTTPStatus.OK:
            return True
        elif (
            response.status_code == HTTPStatus.UNAUTHORIZED
            or response.status_code == HTTPStatus.FORBIDDEN
        ):
            return False
        else:
            response.raise_for_status()

    @staticmethod
    def _get_raise_for_status(response: httpx.Response) -> None:
        # A handler for errors that might be sent with messages from Top.
        if (
            isinstance(response, httpx.Response)
            and not response.status_code == httpx.codes.OK
        ):
            try:
                content = response.json()
            except json.JSONDecodeError:
                response.raise_for_status()

            # Every exception has content of {detail, status_code[, headers]}
            # TODO Some exceptions in Top send detail as a string and not a dict.
            # These exceptions are now handled like normal HTTP exceptions.
            # Need to rewrite these to all have the same format.
            detail = content.pop("detail", "")
            message = None
            # In most cases detail is not a string but a dict.
            if isinstance(detail, dict):
                message = detail.pop("message", None)
            elif not isinstance(detail, str):
                detail = ""
            detail = detail or ""
            # If a message was delivered we want to show that. Otherwise it's a
            # standard HTTP error and should be handled by raise_for_status()
            if message is not None:
                raise Exception(f"Error {response.status_code}: {message} {detail}")
            else:
                response.raise_for_status()

    def upload_file(self, file_or_path) -> FileGet:

        if isinstance(file_or_path, str):
            # TODO: Change this to wrap the file object reader to convert to hex
            # everytime anything is read instead of reading it all at once.

            with open(file_or_path, "rb") as file:
                buffer = file.read()
            hex_buffer = buffer.hex()
            return self._post_file(
                "/v2/files/",
                io.BytesIO(hex_buffer.encode()),
            )
        else:
            return self._post_file("/v2/files/", file_or_path)

    def upload_data(self, file_or_path) -> DataGet:
        uploaded_file = self.upload_file(file_or_path)
        uploaded_data = self._post("/v2/data", uploaded_file.dict())
        return DataGet.parse_obj(uploaded_data)

    def upload_python_object_to_file(self, obj) -> FileGet:
        return self.upload_file(io.BytesIO(python_object_to_hex(obj).encode()))

    def _initialise_direct_pipeline_file_upload(self, file_size: int) -> str:
        """Initialise a direct multi-part pipeline file upload"""
        direct_upload_schema = PipelineFileDirectUploadInitCreate(
            file_size=file_size, file_format=FileFormat.binary
        )
        response = self._post(
            "/v2/pipeline-files/initiate-multipart-upload", direct_upload_schema.dict()
        )
        direct_upload_get = PipelineFileDirectUploadInitGet.parse_obj(response)
        return direct_upload_get.pipeline_file_id

    def _direct_upload_pipeline_file_chunk(
        self,
        data: Union[io.BytesIO, CallbackIOWrapper],
        pipeline_file_id: str,
        part_num: int,
    ) -> MultipartUploadMetadata:
        """Upload a single chunk of a multi-part pipeline file upload.

        Returns the metadata associated with this upload (this is needed to pass into
        the finalisation step).
        """
        # get presigned URL
        part_upload_schema = PipelineFileDirectUploadPartCreate(
            pipeline_file_id=pipeline_file_id, part_num=part_num
        )
        response = self._post(
            "/v2/pipeline-files/presigned-url", part_upload_schema.dict()
        )
        part_upload_get = PipelineFileDirectUploadPartGet.parse_obj(response)
        # upload file chunk
        response = httpx.put(
            part_upload_get.upload_url,
            content=data,
            timeout=self.timeout,
        )

        etag = response.headers["ETag"]
        return MultipartUploadMetadata(ETag=etag, PartNumber=part_num)

    def _finalise_direct_pipeline_file_upload(
        self, pipeline_file_id: str, multipart_metadata: List[MultipartUploadMetadata]
    ) -> PipelineFileGet:
        """Finalise the direct multi-part pipeline file upload"""
        finalise_upload_schema = PipelineFileDirectUploadFinaliseCreate(
            pipeline_file_id=pipeline_file_id,
            multipart_metadata=multipart_metadata,
        )
        response = self._post(
            "/v2/pipeline-files/finalise-multipart-upload",
            finalise_upload_schema.dict(),
        )
        return PipelineFileGet.parse_obj(response)

    def upload_pipeline_file(
        self, pipeline_file: PipelineFile
    ) -> PipelineFileVariableGet:
        """Upload PipelineFile given by pipeline_file.

        Since PipelineFiles can be very large, we implement this slightly
        differently to regular file uploads:
        - We need to split the file into chunks based on FILE_CHUNK_SIZE
        - We first initialise the multi-part upload with the server
        - We then upload the file in chunks (requesting a presigned upload URL for each
            chunk beforehand)
        - Lastly, we finalise the multi-part upload with the server
        """

        file_size = os.path.getsize(pipeline_file.path)

        pipeline_file_id = self._initialise_direct_pipeline_file_upload(
            file_size=file_size
        )

        parts = []
        file_hash = hashlib.sha256()
        if self.verbose:
            progress = tqdm(
                desc=f"{PIPELINE_FILE_STR} Uploading {pipeline_file.path}",
                unit="B",
                unit_scale=True,
                total=file_size,
                unit_divisor=1024,
            )
        with open(pipeline_file.path, "rb") as f:
            while True:
                file_data = f.read(FILE_CHUNK_SIZE)
                if not file_data:
                    if self.verbose:
                        progress.close()
                    break
                file_hash.update(file_data)
                part_num = len(parts) + 1
                # If verbose then wrap our data object in a tqdm callback
                if self.verbose:
                    data = CallbackBytesIO(progress.update, file_data)
                else:
                    data = io.BytesIO(file_data)

                upload_metadata = self._direct_upload_pipeline_file_chunk(
                    data=data,
                    pipeline_file_id=pipeline_file_id,
                    part_num=part_num,
                )
                parts.append(upload_metadata)

        file_hash = file_hash.hexdigest()
        pipeline_file_get = self._finalise_direct_pipeline_file_upload(
            pipeline_file_id=pipeline_file_id, multipart_metadata=parts
        )
        return PipelineFileVariableGet(
            path=pipeline_file.path, file=pipeline_file_get.file, hash=file_hash
        )

    def _get(self, endpoint: str, params: Dict[str, Any] = None):
        headers = {
            "Authorization": "Bearer %s" % self.token,
        }

        url = urllib.parse.urljoin(self.url, endpoint)
        response = httpx.get(url, headers=headers, params=params, timeout=self.timeout)
        response.raise_for_status()
        return response.json()

    def _post(self, endpoint: str, json_data: dict) -> dict:
        self.raise_for_invalid_token()
        headers = {
            "Authorization": "Bearer %s" % self.token,
            "Content-type": "application/json",
        }

        url = urllib.parse.urljoin(self.url, endpoint)
        response = httpx.post(
            url, headers=headers, json=json_data, timeout=self.timeout
        )

        if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
            schema = json_data
            raise InvalidSchema(schema=schema)
        else:
            self._get_raise_for_status(response)

        return response.json()

    def _patch(self, endpoint: str, json_data: dict) -> dict:
        self.raise_for_invalid_token()
        headers = {
            "Authorization": "Bearer %s" % self.token,
            "Content-type": "application/json",
        }

        url = urllib.parse.urljoin(self.url, endpoint)
        response = httpx.patch(
            url, headers=headers, json=json_data, timeout=self.timeout
        )

        if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
            schema = json_data
            raise InvalidSchema(schema=schema)
        else:
            self._get_raise_for_status(response)

        return response.json()

    def _delete(self, endpoint: str) -> None:
        self.raise_for_invalid_token()
        headers = {
            "Authorization": "Bearer %s" % self.token,
            "Content-type": "application/json",
        }

        url = urllib.parse.urljoin(self.url, endpoint)
        response = httpx.delete(url, headers=headers, timeout=self.timeout)
        if response.status_code != HTTPStatus.NO_CONTENT:
            self._get_raise_for_status(response)

    def _post_file(self, endpoint: str, file: io.IOBase) -> FileGet:
        self.raise_for_invalid_token()
        if not hasattr(file, "name"):
            file.name = generate_id(20)

        file_size = file.getbuffer().nbytes

        if self.verbose:
            progress = tqdm(
                desc=f"{PIPELINE_STR} Uploading",
                unit="B",
                unit_scale=True,
                total=file_size,
                unit_divisor=1024,
            )
            # If verbose then wrap our file object in a tqdm callback
            file = CallbackIOWrapper(progress.update, file)

        headers = {
            "Authorization": "Bearer %s" % self.token,
        }
        url = urllib.parse.urljoin(self.url, endpoint)
        response = httpx.post(
            url,
            headers=headers,
            files={"file": (file.name, file, "application/octet-stream")},
            timeout=self.timeout,
        )
        if self.verbose:
            progress.close()
        if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
            schema = FileCreate.__name__
            raise InvalidSchema(schema=schema)
        else:
            response.raise_for_status()
        return FileGet.parse_obj(response.json())

    def upload_function(self, function: Function) -> FunctionGet:
        try:
            inputs = [
                dict(name=name, type_name=python_object_to_name(type))
                for name, type in function.typing_inputs.items()
            ]
            output = [
                dict(name=name, type_name=python_object_to_name(type))
                for name, type in function.typing_outputs.items()
            ]

            file_schema = self.upload_python_object_to_file(function)

            function_create_schema = FunctionCreate(
                local_id=function.local_id,
                name=function.name,
                function_source=function.source,
                hash=function.hash,
                inputs=inputs,
                output=output,
                file_id=file_schema.id,
            )
        except AttributeError as e:
            raise InvalidSchema(schema="Function", message=str(e))

        response = self._post("/v2/functions", function_create_schema.dict())
        return FunctionGet.parse_obj(response)

    def upload_model(self, model: Model) -> ModelGet:
        file_schema = self.upload_python_object_to_file(model)
        try:
            model_create_schema = ModelCreate(
                local_id=model.local_id,
                name=model.name,
                model_source=model.source,
                hash=model.hash,
                file_id=file_schema.id,
            )
        except ValidationError as e:
            raise InvalidSchema(schema="Model", message=str(e))

        response = self._post("/v2/models", model_create_schema.dict())
        return ModelGet.parse_obj(response)

    def upload_pipeline(
        self,
        new_pipeline_graph: Graph,
        public: bool = False,
        description: str = "",
        tags: Set[str] = None,
    ) -> PipelineGet:
        """
        Upload a Pipeline to the Cloud.

            Parameters:
                    new_pipeline_graph (Graph): Graph repr. Pipeline to be uploaded.
                        Obtained from Pipeline.get_pipeline(name:str)
                    public (bool): If pipeline should be visible to public.
                        Defaults to False.
                    description (str): Description of the Pipeline.
                    tags (Set[str]): Set of tags for the pipeline. Eg: {"BERT", "NLP"}

            Returns:
                    pipeline (PipelineGet): Object representing uploaded pipeline.
        """
        if new_pipeline_graph._has_run_startup:
            raise Exception("Cannot upload a pipeline that has already been run.")
        # Pipeline Cloud currently supports Python 3.9.x only
        if (sys.version_info.major, sys.version_info.minor) != (3, 9):
            print(
                "WARNING: pipeline-ai is still in development and the"
                " upload_pipeline function has only been tested in Python 3.9. "
                "We strongly recommend you use Python 3.9 as pipelines uploaded"
                " in other Python versions are known to be broken. We are working"
                "on adding support for 3.10 and 3.8!"
            )

        new_name = new_pipeline_graph.name
        if self.verbose:
            print("Uploading functions")
        new_functions = [
            self.upload_function(_function)
            for _function in new_pipeline_graph.functions
        ]
        for i, uploaded_function_schema in enumerate(new_functions):
            new_pipeline_graph.functions[i].local_id = uploaded_function_schema.id
        if self.verbose:
            print("Uploading models")
        new_models = [self.upload_model(_model) for _model in new_pipeline_graph.models]

        new_variables: List[PipelineVariableGet] = []
        if self.verbose:
            print("Uploading variables")

        from pipeline.objects import PipelineFile

        for _var in new_pipeline_graph.variables:
            _var_type_file = self.upload_file(
                io.BytesIO(python_object_to_hex(_var.type_class).encode())
            )

            pipeline_file_schema = None
            if isinstance(_var, PipelineFile):
                if _var.remote_id is not None:
                    file_schema: FileGet = self._download_schema(
                        schema=FileGet,
                        endpoint=f"/v2/files/{_var.remote_id}",
                        params=dict(
                            return_data=False,
                        ),
                    )
                    unique_identifier = str(uuid.uuid4())
                    pipeline_file_schema = PipelineFileVariableGet(
                        path=unique_identifier, hash=unique_identifier, file=file_schema
                    )
                else:
                    pipeline_file_schema = self.upload_pipeline_file(_var)

            _var_schema = PipelineVariableGet(
                local_id=_var.local_id,
                name=_var.name,
                type_file=_var_type_file,
                is_input=_var.is_input,
                is_output=_var.is_output,
                pipeline_file_variable=pipeline_file_schema,
            )

            new_variables.append(_var_schema)

        new_graph_nodes = [
            _node.to_create_schema() for _node in new_pipeline_graph.nodes
        ]
        new_outputs = [_output.local_id for _output in new_pipeline_graph.outputs]

        compute_requirements = None
        if new_pipeline_graph.min_gpu_vram_mb:
            compute_requirements = ComputeRequirements(
                min_gpu_vram_mb=new_pipeline_graph.min_gpu_vram_mb
            )

        try:
            pipeline_create_schema = PipelineCreate(
                name=new_name,
                variables=new_variables,
                functions=new_functions,
                models=new_models,
                graph_nodes=new_graph_nodes,
                outputs=new_outputs,
                public=public,
                description=description,
                tags=tags or set(),
                compute_type=new_pipeline_graph.compute_type,
                compute_requirements=compute_requirements,
            )
        except ValidationError as e:
            raise InvalidSchema(schema="Graph", message=str(e))

        if self.verbose:
            print("Uploading pipeline graph")
        response = self._post(
            "/v2/pipelines", json.loads(pipeline_create_schema.json())
        )
        return PipelineGet.parse_obj(response)

    def run_pipeline(
        self,
        pipeline_id_or_schema: Union[str, PipelineGet],
        raw_data_or_schema: Union[Any, DataGet],
        compute_type: Optional[str] = None,
        min_gpu_vram_mb: Optional[int] = None,
    ):
        """
        Uploads Data and executes a Run of given pipeline over given data.

            Parameters:
                    pipeline_id_or_schema (Union[str, PipelineGet]):
                        The id for the desired pipeline
                        or the schema obtained from uploading it
                    raw_data_or_schema (Union[Any, DataGet]):
                        Raw data for Pipeline execution
                        or schema for already uploaded data.
                    compute_type (Optional[str]):
                        Compute type requirement for this run (e.g. "cpu" or "gpu"). If
                        not defined, it will fallback to that specified by the Pipeline
                        itself (typically this is "gpu")
                    min_gpu_vram_mb (Optional[int]):
                        Minimum amount of GPU VRAM required. If not defined, it will
                        fallback to that specified by the Pipeline (if applicable). This
                        should only be specified for GPU workloads.

            Returns:
                    run (Any): Run object containing metadata and outputs.
        """
        # TODO: Add support for generic object inference. Only strs at the moment.
        if not isinstance(raw_data_or_schema, DataGet):
            temp_file = io.BytesIO(python_object_to_hex(raw_data_or_schema).encode())
            uploaded_data = self.upload_data(temp_file)
            _data_id = uploaded_data.id
        elif isinstance(raw_data_or_schema, DataGet):
            _data_id = raw_data_or_schema.id
        else:
            raise Exception("Must either pass a raw data, or DataGet schema.")

        pipeline_id = None
        if isinstance(pipeline_id_or_schema, str):
            pipeline_id = pipeline_id_or_schema
        elif isinstance(pipeline_id_or_schema, PipelineGet):
            pipeline_id = pipeline_id_or_schema.id
        else:
            raise InvalidSchema(
                schema=pipeline_id_or_schema,
                message=(
                    "Must either pass a pipeline id, or a pipeline get schema. "
                    "Not object of type %s in arg 1." % str(pipeline_id_or_schema)
                ),
            )

        compute_requirements = None
        if min_gpu_vram_mb:
            compute_requirements = ComputeRequirements(min_gpu_vram_mb=min_gpu_vram_mb)

        run_create_schema = RunCreate(
            pipeline_id=pipeline_id,
            data_id=_data_id,
            compute_type=compute_type,
            compute_requirements=compute_requirements,
        )
        run_json: dict = self._post("/v2/runs", json.loads(run_create_schema.json()))
        return RunGet.parse_obj(run_json)

    def _download_schema(
        self, schema: Type[BaseModel], endpoint: str, params: Optional[Dict[str, Any]]
    ) -> Type[BaseModel]:
        """
        Request json data to a given endpoint and parse it into given schema

            Parameters:
                schema (Type[BaseModel]): Which schema is expected to be returned
                endpoint (str): endpoint to which the request must be sent
                params (Optional[Dict[str, Any]]): optional request params

            Returns:
                schema (Type[BaseModel]): The populated schema passed in
        """
        response = self._get(endpoint=endpoint, params=params)
        try:
            return schema(**response)
        except ValidationError as e:
            raise InvalidSchema(schema=schema.__name__, message=str(e))

    def download_function(self, id: str) -> Function:
        """
        Downloads Function object from Pipeline Cloud.

            Parameters:
                    id (str):
                        The id for the desired function

            Returns:
                    function (Function): De-Serialized function.
        """
        endpoint = f"/v2/functions/{id}"
        f_get_schema: FunctionGet = self._download_schema(
            schema=FunctionGet,
            endpoint=endpoint,
            params=dict(return_data=True),
        )
        # FIXME we have a circular import issue that needs reviewing
        from pipeline.objects import Function

        return Function.from_schema(f_get_schema)

    def download_model(self, id: str) -> Model:
        """
        Downloads Model object from Pipeline Cloud.

            Parameters:
                    id (str):
                        The id for the desired model

            Returns:
                    model (Model): De-Serialized model.
        """
        endpoint = f"/v2/models/{id}"
        m_get_schema: ModelGet = self._download_schema(
            schema=ModelGet,
            endpoint=endpoint,
            params=dict(return_data=True),
        )
        # FIXME we have a circular import issue that needs reviewing
        from pipeline.objects import Model

        return Model.from_schema(m_get_schema)

    def download_data(self, id: str) -> Any:
        """
        Downloads Data object from Pipeline Cloud.

            Parameters:
                    id (str):
                        The id for the desired data

            Returns:
                    object (Any): De-Serialized data object.
        """
        endpoint = f"/v2/data/{id}"
        d_get_schema: DataGet = self._download_schema(
            schema=DataGet,
            endpoint=endpoint,
            params=dict(return_data=True),
        )
        return hex_to_python_object(d_get_schema.hex_file.data)

    def download_result(self, result_id_or_schema: Union[str, RunGet]) -> Any:
        """
        Downloads Result object from Pipeline Cloud.
            Parameters:
                    result_id_or_schema (Union[str, RunGet]):
                    The id for the desired run result
                    or the schema obtained from the run

            Returns:
                    object (Any): De-Serialized run result file object.
        """
        result_id = None
        if isinstance(result_id_or_schema, str):
            result_id = result_id_or_schema
        else:
            try:
                result_id = RunGet.parse_obj(result_id_or_schema).result.id
            except ValidationError:
                raise InvalidSchema(
                    schema=result_id_or_schema,
                    message=(
                        "Must either pass a result id, or a run get schema. "
                        "Not object of type %s in arg 1." % str(result_id_or_schema)
                    ),
                )
        endpoint = f"/v2/files/{result_id}"
        f_get_schema: FileGet = self._download_schema(
            schema=FileGet, endpoint=endpoint, params=dict(return_data=True)
        )
        return hex_to_python_object(f_get_schema.data)

    def download_pipeline(self, id: str) -> Graph:
        """
        Downloads Graph object from Pipeline Cloud.

            Parameters:
                    id (str):
                        The id for the desired pipeline

            Returns:
                    graph (Graph): De-Serialized pipeline.
        """
        endpoint = f"/v2/pipelines/{id}"
        p_get_schema: PipelineGet = self._download_schema(
            schema=PipelineGet,
            endpoint=endpoint,
            params=dict(return_data=True),
        )
        # FIXME we have a circular import issue that needs reviewing
        from pipeline.objects import Graph

        return Graph.from_schema(p_get_schema)

    def get_runs(
        self, limit: int = 20, skip: int = 0, created_at_order: str = "desc"
    ) -> List[RunGet]:
        # Get runs from the remote compute service
        result = self._get(
            "/v2/runs",
            params=dict(
                limit=limit, skip=skip, order_by=f"created_at:{created_at_order}"
            ),
        )
        return result

    def download_remotes(self, graph: Graph) -> None:
        # Only remote PipelineFiles are supported
        for variable in graph.variables:
            if not isinstance(variable, PipelineFile) or variable.remote_id is None:
                continue

            downloaded_schema: FileGet = self._download_schema(
                schema=FileGet,
                endpoint=f"/v2/files/{variable.remote_id}",
                params=dict(
                    return_data=True,
                ),
            )

            configuration.PIPELINE_CACHE_FILES.mkdir(exist_ok=True)
            raw_result = hex_to_python_object(downloaded_schema.data)
            file_path = configuration.PIPELINE_CACHE_FILES / variable.remote_id
            with open(file_path, "wb") as file:
                dill.dump(raw_result, file=file)

            variable.path = file_path



from datetime import datetime
from typing import Dict, List, Optional, Set

from pydantic import Field, root_validator, validator

from pipeline.schemas.base import BaseModel
from pipeline.schemas.compute_requirements import ComputeRequirements, ComputeType
from pipeline.schemas.file import FileGet
from pipeline.schemas.function import FunctionGet
from pipeline.schemas.model import ModelGet
from pipeline.schemas.runnable import RunnableGet, RunnableType


class PipelineGraphNode(BaseModel):
    local_id: str
    function: str
    inputs: List[str]
    outputs: List[str]


class PipelineFileVariableGet(BaseModel):
    path: str
    hash: str
    file: FileGet


class PipelineVariableGet(BaseModel):
    local_id: str
    name: Optional[str]

    type_file: Optional[FileGet]
    type_file_id: Optional[str]

    is_input: bool
    is_output: bool

    pipeline_file_variable: Optional[PipelineFileVariableGet]

    @root_validator
    def file_or_id_validation(cls, values):
        file, file_id = values.get("type_file"), values.get("type_file_id")

        file_defined = file is not None
        file_id_defined = file_id is not None

        if file_defined == file_id_defined:
            raise ValueError(
                "You must define either the type_file OR type_file_id of a variable."
            )

        return values


class PipelineGetBrief(BaseModel):
    id: str
    name: str
    deployed: bool = False
    tags: List[str] = []
    description: str = ""


class PipelineGet(PipelineGetBrief, RunnableGet):
    type: RunnableType = Field(RunnableType.pipeline, const=True)
    variables: List[PipelineVariableGet]
    functions: List[FunctionGet]
    models: List[ModelGet]
    graph_nodes: List[PipelineGraphNode]
    outputs: List[str]

    class Config:
        orm_mode = True


class PipelineGetDetailed(PipelineGet):
    version: str = "1"
    dependencies: List[str] = ["torch", "transformers"]
    created_at: datetime
    updated_at: datetime
    last_run: Optional[datetime]
    public: bool
    # Maps language, e.g. `curl` or `python`, to an example Run creation code snippet
    run_examples: Dict[str, str] = {}


class PipelineCreate(BaseModel):
    name: str
    variables: List[PipelineVariableGet]
    functions: List[FunctionGet]
    models: List[ModelGet]
    graph_nodes: List[PipelineGraphNode]
    outputs: List[str]
    public: bool = False
    description: str = ""
    tags: Set[str] = set()
    # By default a Pipeline will require GPU resources
    compute_type: ComputeType = ComputeType.gpu
    compute_requirements: Optional[ComputeRequirements]

    @validator("compute_requirements")
    def compute_type_is_gpu(cls, v, values):
        """If compute_type is not 'gpu' we don't expect any additional compute
        requirements to be specified.
        """
        if values["compute_type"] != ComputeType.gpu:
            if v and v.min_gpu_vram_mb:
                raise ValueError(
                    "min_gpu_vram_mb should only be specified for gpu workloads"
                )
        return v


class PipelineTagCreate(BaseModel):
    # The full name of the tag, e.g. `my-pipeline:latest`.
    name: str
    # The pipeline ID this tag should point to.
    pipeline_id: str


class PipelineTagGet(BaseModel):
    id: str
    name: str
    pipeline_id: str


class PipelineTagPatch(BaseModel):
    # The new pipeline ID this tag should point to.
    pipeline_id: str



from typing import List

from pipeline.objects.function import Function
from pipeline.objects.variable import Variable
from pipeline.schemas.pipeline import PipelineGraphNode
from pipeline.util import generate_id


class GraphNode:
    local_id: str
    function: Function
    inputs: List[Variable] = []
    outputs: List[Variable] = []

    def __init__(self, function, inputs, outputs, *, local_id=None):
        self.function = function
        self.inputs = inputs
        self.outputs = outputs

        self.local_id = generate_id(10) if local_id is None else local_id

    def to_create_schema(self) -> PipelineGraphNode:
        return PipelineGraphNode(
            local_id=self.local_id,
            function=self.function.local_id,
            inputs=[_var.local_id for _var in self.inputs],
            outputs=[_var.local_id for _var in self.outputs],
        )



from typing import Optional

from .base import BaseModel


class ResourceBase(BaseModel):
    foreign_id: Optional[str]
    resource_label: str
    resource_type: str


class ResourceCreate(ResourceBase):
    pass


class ResourceGet(ResourceBase):
    id: str



from transformers import AutoModelForCausalLM, AutoTokenizer

from pipeline import pipeline_function, pipeline_model


@pipeline_model
class TransformersModelForCausalLM:
    def __init__(
        self,
        model_path: str = "EleutherAI/gpt-neo-125M",
        tokenizer_path: str = "EleutherAI/gpt-neo-125M",
    ):
        self.model_path = model_path
        self.tokenizer_path = tokenizer_path
        self.model = None
        self.tokenizer = None

    @pipeline_function
    def predict(self, input_data: str, model_kwargs: dict) -> str:

        input_ids = self.tokenizer(input_data, return_tensors="pt").input_ids
        gen_tokens = self.model.generate(input_ids, **model_kwargs)
        gen_text = self.tokenizer.batch_decode(gen_tokens)[0]
        return gen_text

    @pipeline_function(on_startup=True, run_once=True)
    def load(self) -> None:
        if self.model is None:
            self.model = AutoModelForCausalLM.from_pretrained(self.model_path)
        if self.tokenizer is None:
            self.tokenizer = AutoTokenizer.from_pretrained(self.tokenizer_path)



from typing import Optional

from pipeline.schemas.base import BaseModel
from pipeline.schemas.pipeline import PipelineGet


class DeploymentCreate(BaseModel):
    """Schema for creating a pipeline deployment"""

    pipeline_id: str


class DeploymentGet(BaseModel):
    id: str
    pipeline: PipelineGet
    active: bool


class DeploymentPatch(BaseModel):
    active: Optional[bool]



from typing import List, Optional

from .base import BaseModel
from .file import FileFormat, FileGet


class PipelineFileDirectUploadInitCreate(BaseModel):
    file_size: int
    # hex is the default purely for backwards-compatability
    file_format: FileFormat = FileFormat.hex


class PipelineFileDirectUploadInitGet(BaseModel):
    pipeline_file_id: str


class PipelineFileDirectUploadPartCreate(BaseModel):
    pipeline_file_id: str
    # The part number for this multi-part file upload
    part_num: int


class PipelineFileDirectUploadPartGet(BaseModel):
    # The URL to use when uploading the fie
    upload_url: str


class MultipartUploadMetadata(BaseModel):
    """Schema for multi-part uploads direct to storage server.

    (We don't have any control over this schema)
    """

    ETag: str
    PartNumber: int


class PipelineFileDirectUploadFinaliseCreate(BaseModel):
    pipeline_file_id: str
    # The metadata obtained from each part of the file upload
    multipart_metadata: List[MultipartUploadMetadata]


class PipelineFileGet(BaseModel):
    id: str
    name: str
    # This is a legacy field that is replaced by 'file'.
    # Can be deprecated once no clients are using it.
    hex_file: Optional[FileGet]
    file: Optional[FileGet]



import base64
import json
import os
import sys
from pathlib import Path
from platform import python_version
from typing import TypedDict

from packaging import version

from pipeline.util.logging import _print

PIPELINE_CACHE = Path(
    os.getenv(
        "PIPELINE_CACHE",
        Path(os.getenv("LOCALAPPDATA")) / ".pipeline/"
        if (sys.platform == "win32" or sys.platform == "cygwin")
        else Path.home() / ".cache/pipeline/",
    )
)

PIPELINE_CACHE_FILES = PIPELINE_CACHE / "files"
PIPELINE_CACHE_CONFIG = PIPELINE_CACHE / "config.json"
PIPELINE_CACHE_AUTH = PIPELINE_CACHE / "auth.json"

DEFAULT_REMOTE: str = None

if version.parse(python_version()) < version.parse("3.9.13"):
    _print(
        f"You are using python version '{python_version()}' "
        "please upgrade to python >=3.9.13 to ensure correct serialisation.",
        "WARNING",
    )

remote_auth: TypedDict("remote_auth", {"url": str}) = dict()
config: dict = dict()


def _load_auth():
    global remote_auth
    if PIPELINE_CACHE_AUTH.exists():
        remote_auth = json.loads(PIPELINE_CACHE_AUTH.read_text())
        remote_auth = {
            url: base64.b64decode(encoded_token).decode()
            for url, encoded_token in remote_auth.items()
        }


def _save_auth():
    PIPELINE_CACHE.mkdir(exist_ok=True)

    with open(PIPELINE_CACHE_AUTH, "w") as auth_file:

        _b64_remote_auth = {
            auth[0]: base64.b64encode(auth[1].encode()).decode()
            for auth in remote_auth.items()
        }

        auth_file.write(json.dumps(_b64_remote_auth))


def _load_config():
    global DEFAULT_REMOTE
    global config
    if PIPELINE_CACHE_CONFIG.exists():
        config = json.loads(PIPELINE_CACHE_CONFIG.read_text())

    DEFAULT_REMOTE = config.get("DEFAULT_REMOTE", "https://api.pipeline.ai")


def _save_config():
    global config

    PIPELINE_CACHE.mkdir(exist_ok=True)

    with open(PIPELINE_CACHE_CONFIG, "w") as config_file:
        config_file.write(json.dumps(config))


_load_config()
_load_auth()



import argparse
import sys
from typing import List, Optional

from pipeline.console.remote import remote as remote_command
from pipeline.console.runs import runs as runs_command
from pipeline.console.tags import tags as tags_command


def main(args: Optional[List[str]] = None) -> int:
    base_parser = argparse.ArgumentParser(
        prog="pipeline",
        description="Create or run pipelines locally or in the cloud!",
        add_help=True,
    )
    base_parser.add_argument(
        "-v",
        "--verbose",
        help="Verbose logging",
        action="store_true",
    )

    command_parser = base_parser.add_subparsers(dest="command")
    ##########
    # pipeline remote
    ##########

    remote_parser = command_parser.add_parser(
        "remote",
        description="Manage remote compute services",
        help="Manage remote compute services",
    )
    remote_sub_parser = remote_parser.add_subparsers(dest="sub-command")

    ##########
    # pipeline remote login
    ##########
    remote_login_parser = remote_sub_parser.add_parser(
        "login",
        description="Authenticate with a remote compute service",
        help="Authenticate with a remote compute service",
    )

    remote_login_parser.add_argument(
        "-u",
        "--url",
        type=str,
        required=False,
        help="Remote URL for auth (default=https://api.pipeline.ai)",
        # description="Remote URL for auth (default=https://api.pipeline.ai)",
        default="https://api.pipeline.ai",
    )
    remote_login_parser.add_argument(
        "-t",
        "--token",
        type=str,
        required=True,
        help="API token for remote",
    )

    ##########
    # pipeline remote set
    ##########

    remote_set_parser = remote_sub_parser.add_parser(
        "set",
        description="Set the currently used remote compute service URL",
        help="Set the currently used remote compute service URL",
    )
    remote_set_parser.add_argument("url", help="The remote URL")

    ##########
    # pipeline remote list
    ##########

    remote_sub_parser.add_parser(
        "list",
        aliases=["ls"],
        description="List the authenticated remote compute services",
        help="List the authenticated remote compute services",
    )

    ##########
    # pipeline runs
    ##########

    runs_parser = command_parser.add_parser(
        "runs",
        description="Manage pipeline runs",
        help="Manage pipeline runs",
    )
    runs_sub_parser = runs_parser.add_subparsers(dest="sub-command")

    ##########
    # pipeline runs list
    ##########

    runs_sub_parser.add_parser(
        "list",
        aliases=["ls"],
        description="List the currently executing runs",
        help="List the currently executing runs",
    )
    ##########
    # pipeline runs get
    ##########

    runs_get_parser = runs_sub_parser.add_parser(
        "get",
        description="Get run information from a remote compute service",
        help="Get run information from a remote compute service",
    )

    runs_get_parser.add_argument(
        "run_id",
        help="The run id",
    )

    runs_get_parser.add_argument(
        "-r",
        "--result",
        action="store_true",
        help="Get the run result",
    )

    ##########
    # pipeline tags
    ##########

    tags_parser = command_parser.add_parser(
        "tags",
        description="Manage pipeline tags",
        help="Manage pipeline tags",
    )

    tags_sub_parser = tags_parser.add_subparsers(dest="sub-command")

    ##########
    # pipeline tags create
    ##########

    tags_set_parser = tags_sub_parser.add_parser(
        "create",
        help="Create a tag TARGET that points to SOURCE",
    )

    tags_set_parser.add_argument(
        "source",
        help="The source pipeline:tag or pipeline_id",
    )
    tags_set_parser.add_argument(
        "target",
        help="The target pipeline:tag",
    )

    ##########
    # pipeline tags update
    ##########

    tags_update_parser = tags_sub_parser.add_parser(
        "update",
        help="Update a tag TARGET to point to a new SOURCE",
    )
    tags_update_parser.add_argument(
        "source",
        help="The source pipeline:tag or pipeline_id",
    )
    tags_update_parser.add_argument(
        "target",
        help="The target pipeline:tag",
    )

    ##########
    # pipeline tags list
    ##########

    tags_list_parser = tags_sub_parser.add_parser(
        "list",
        aliases=["ls"],
        help="List tags on the remote compute service",
    )

    tags_list_parser.add_argument(
        "-p",
        "--pipeline-id",
        type=str,
        help="Filter by target pipeline id",
    )
    tags_list_parser.add_argument(
        "-l",
        "--limit",
        required=False,
        help="Number of tags to get",
        default=20,
        type=int,
    )
    tags_list_parser.add_argument(
        "-s",
        "--skip",
        required=False,
        help="Number of tags to skip for pagination",
        default=0,
        type=int,
    )

    ##########
    # pipeline tags delete
    ##########

    tags_delete_parser = tags_sub_parser.add_parser(
        "delete",
        aliases=["rm"],
        help="Delete a pipeline tag (does not delete the pipeline)",
    )

    tags_delete_parser.add_argument(
        "pipeline_tag", help="The pipeline tag or tag_id to delete"
    )

    ##########
    # pipeline tags get
    ##########

    tags_get_parser = tags_sub_parser.add_parser(
        "get",
        help="Get tag information",
    )

    tags_get_parser.add_argument(
        "pipeline_tag", help="The pipeline tag or tag_id to get"
    )

    ##########
    args: argparse.Namespace = base_parser.parse_args(args)
    command = getattr(args, "command", None)

    if command == "remote":
        if (code := remote_command(args)) is None:
            remote_parser.print_help()
            return 1
    elif command == "runs":
        if (code := runs_command(args)) is None:
            runs_parser.print_help()
            return 1
    elif command == "tags":
        if (code := tags_command(args)) is None:
            tags_parser.print_help()
            return 1
    else:
        base_parser.print_help()
        return 0

    return code


def _run() -> int:
    sys.exit(main())



from typing import Generic, List, TypeVar

from .base import BaseModel, GenericModel

DataType = TypeVar("DataType")


class PaginationDetails(BaseModel):
    """Query parameters for requesting paginated resource lists."""

    #: Index of resource items to begin paginating from
    skip: int
    #: Maximum number of resources to return
    limit: int


class Paginated(GenericModel, Generic[DataType]):
    """Response for paginated resource lists."""

    #: Index of resource items to begin paginating from
    skip: int
    #: Maximum number of resources the `data` field will contain
    limit: int
    #: Total number of resources available for pagination
    total: int
    #: Resource data
    data: List[DataType]

    @classmethod
    def of(cls, item_list: List[DataType], details: PaginationDetails, total: int):

        return Paginated(
            skip=details.skip, limit=details.limit, total=total, data=item_list
        )



class InvalidSchema(Exception):
    def __init__(self, schema=None, message="Invalid Schema") -> None:
        self.schema = schema
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.schema} -> {self.message}"



from pipeline.objects.decorators import pipeline_function, pipeline_model
from pipeline.objects.function import Function
from pipeline.objects.graph import Graph
from pipeline.objects.model import Model
from pipeline.objects.pipeline import Pipeline
from pipeline.objects.variable import PipelineFile, Variable
from pipeline.objects.wrappers import onnx_to_pipeline

__all__ = [
    "Pipeline",
    "Graph",
    "Function",
    "Model",
    "Variable",
    "pipeline_function",
    "pipeline_model",
    "PipelineFile",
    "onnx_to_pipeline",
]



from enum import Enum
from typing import Optional

from .base import BaseModel


# https://github.com/samuelcolvin/pydantic/issues/2278
class ComputeType(str, Enum):
    cpu: str = "cpu"
    gpu: str = "gpu"


class ComputeRequirements(BaseModel):
    """Additional compute requirements required by a runnable (e.g. min GPU VRAM)"""

    min_gpu_vram_mb: Optional[int]



from typing import Any

from pipeline.schemas.pipeline import PipelineVariableGet
from pipeline.util import generate_id, hex_to_python_object


class Variable:

    local_id: str
    remote_id: str

    name: str

    type_class: Any

    is_input: bool
    is_output: bool

    def __init__(
        self,
        type_class: Any,
        *,
        is_input: bool = False,
        is_output: bool = False,
        name: str = None,
        remote_id: str = None,
        local_id: str = None,
    ):
        self.remote_id = remote_id
        self.name = name
        self.type_class = type_class
        self.is_input = is_input
        self.is_output = is_output

        self.local_id = generate_id(10) if not local_id else local_id

    @classmethod
    def from_schema(cls, schema: PipelineVariableGet):
        if schema.pipeline_file_variable is not None:
            return PipelineFile.from_schema(schema)
        else:
            return cls(
                hex_to_python_object(schema.type_file.data),
                is_input=schema.is_input,
                is_output=schema.is_output,
                name=schema.name,
                local_id=schema.local_id,
            )


class PipelineFile(Variable):

    path: str

    def __init__(
        self,
        *,
        path: str = None,
        name: str = None,
        remote_id: str = None,
        local_id: str = None,
    ) -> None:
        super().__init__(
            type_class=self.__class__,
            is_input=False,
            is_output=False,
            name=name,
            remote_id=remote_id,
            local_id=local_id,
        )
        self.path = path

    @classmethod
    def from_schema(cls, schema: PipelineVariableGet):
        return cls(
            path=schema.pipeline_file_variable.path,
            name=schema.name,
            local_id=schema.local_id,
        )



import argparse
import re
import sys
from typing import List

from tabulate import tabulate

from pipeline import PipelineCloud
from pipeline.schemas.pagination import Paginated
from pipeline.schemas.pipeline import (
    PipelineTagCreate,
    PipelineTagGet,
    PipelineTagPatch,
)
from pipeline.util.logging import _print

tag_re_pattern = re.compile(
    r"^[a-z0-9][a-z0-9-._/]*[a-z0-9]:[0-9A-Za-z_][0-9A-Za-z-_.]{0,127}$", re.IGNORECASE
)


def _get_tag(tag_name: str) -> PipelineTagGet:
    if not tag_re_pattern.match(tag_name):
        _print("Source tag must match pattern 'pipeline:tag'", level="ERROR")
        raise sys.exit(1)

    remote_service = PipelineCloud(verbose=False)
    remote_service.authenticate()
    tag_information = PipelineTagGet.parse_obj(
        remote_service._get(f"/v2/pipeline-tags/by-name/{tag_name}")
    )
    return tag_information


def _update_or_create_tag(source: str, target: str, sub_command: str) -> PipelineTagGet:
    remote_service = PipelineCloud(verbose=False)
    remote_service.authenticate()
    if not tag_re_pattern.match(target):
        _print("Target tag must match pattern 'pipeline:tag'", level="ERROR")
        raise sys.exit(1)

    if tag_re_pattern.match(source):
        # Pointing to another tag
        source_schema = _get_tag(source)
        source_pipeline = source_schema.pipeline_id
    else:
        # Pointing to a pipeline_id
        source_pipeline = source

    if sub_command == "create":
        tag_create_schema = PipelineTagCreate(
            name=target,
            pipeline_id=source_pipeline,
        )
        response = remote_service._post(
            "/v2/pipeline-tags", json_data=tag_create_schema.dict()
        )
    else:
        # Update
        existing_tag = _get_tag(target)
        patch_schema = PipelineTagPatch(pipeline_id=source_pipeline)
        response = remote_service._patch(
            f"/v2/pipeline-tags/{existing_tag.id}", json_data=patch_schema.dict()
        )

    tag_get_schema = PipelineTagGet.parse_obj(response)
    return tag_get_schema


def _list_tags(
    skip: int, limit: int, pipeline_id: str = None
) -> Paginated[PipelineTagGet]:
    remote_service = PipelineCloud(verbose=False)
    remote_service.authenticate()
    response = remote_service._get(
        "/v2/pipeline-tags",
        params=dict(
            skip=skip,
            limit=limit,
            order_by="created_at:desc",
            pipeline_id=pipeline_id,
        ),
    )

    paginated_results = Paginated[PipelineTagGet].parse_obj(response)

    return paginated_results


def _delete_tag(tag_name: str) -> None:
    remote_service = PipelineCloud(verbose=False)
    remote_service.authenticate()

    tag_information = _get_tag(tag_name)
    remote_service._delete(f"/v2/pipeline-tags/{tag_information.id}")


def _tabulate_tags(tags: List[PipelineTagGet]) -> str:
    return tabulate(
        [
            [
                tag.id,
                tag.name,
                tag.pipeline_id,
            ]
            for tag in tags
        ],
        headers=[
            "ID",
            "Name",
            "Target",
        ],
        tablefmt="outline",
    )


def tags(args: argparse.Namespace) -> int:
    sub_command = getattr(args, "sub-command", None)

    if sub_command in ["create", "update"]:
        source = getattr(args, "source")
        target = getattr(args, "target")
        tag_get_schema = _update_or_create_tag(source, target, sub_command)
        _print(f"Tag '{tag_get_schema.name}' -> '{tag_get_schema.pipeline_id}'")
        return 0
    elif sub_command in ["list", "ls"]:
        pipeline_id: str = getattr(args, "pipeline_id", None)
        paginated_results = _list_tags(
            getattr(args, "skip"),
            getattr(args, "limit"),
            pipeline_id=pipeline_id,
        )
        table_string = _tabulate_tags(paginated_results.data)
        print(table_string)
        return 0
    elif sub_command in ["delete", "rm"]:
        tag_name = getattr(args, "pipeline_tag")
        _delete_tag(tag_name)
        _print(f"Deleted {tag_name}")
        return 0
    elif sub_command == "get":
        tag_name = getattr(args, "pipeline_tag")
        tag_information = _get_tag(tag_name)
        table_string = _tabulate_tags([tag_information])
        print(table_string)
        return 0



from typing import Optional

from pydantic import root_validator

from pipeline.schemas.base import BaseModel
from pipeline.schemas.file import FileCreate, FileGet


class ModelBase(BaseModel):
    id: str
    name: str


class ModelGet(ModelBase):
    hex_file: FileGet

    source_sample: str

    class Config:
        orm_mode = True


class ModelGetDetailed(ModelGet):
    ...


class ModelGetOverview(ModelBase):
    description: str
    pipeline_count: int


class ModelCreate(BaseModel):
    # The local ID is assigned when a new model is used as part of a new
    # pipeline; the server uses the local ID to associated a model to a
    # Pipeline before replacing the local ID with the server-generated one
    local_id: Optional[str]

    model_source: str
    hash: str
    name: str

    file_id: Optional[str]
    file: Optional[FileCreate]

    @root_validator
    def file_or_id_validation(cls, values):
        file, file_id = values.get("file"), values.get("file_id")

        file_defined = file is not None
        file_id_defined = file_id is not None

        if file_defined == file_id_defined:
            raise ValueError("You must define either the file OR file_id of a model.")

        return values



import copy
import os
import subprocess
from pathlib import Path
from typing import List, Optional

import yaml

from pipeline.objects import Graph, PipelineFile
from pipeline.objects.environment import Environment


def create_pipeline_api(
    pipeline_graphs: List[Graph],
    *,
    output_dir: str = "./",
    platform: str = "linux/amd64",
    environment: Optional[Environment] = None,
    gpu_index: Optional[str] = None,
    **environment_variables,
):
    paths = []
    Path(output_dir).mkdir(parents=True, exist_ok=True)

    pipeline_graphs = copy.deepcopy(pipeline_graphs)

    local_file_paths: List[str] = []

    for pipeline_graph in pipeline_graphs:
        for _var in pipeline_graph.variables:
            if isinstance(_var, PipelineFile):
                local_file_paths.append(_var.path)
                _var.path = os.path.join("/app/files/", os.path.basename(_var.path))

        graph_path = os.path.join(output_dir, pipeline_graph.name + ".graph")
        pipeline_graph.save(graph_path)
        paths.append(graph_path)
    if environment is not None:
        environment.to_requirements()
    create_dockerfile(
        paths,
        output_dir=output_dir,
        platform=platform,
        requirements="requirements.txt",
        pipeline_file_paths=local_file_paths,
    )

    create_docker_compose(output_dir, gpu_index=gpu_index, **environment_variables)


def create_dockerfile(
    pipeline_graph_paths: List[str],
    *,
    output_dir: str = "./",
    platform: str = "linux/amd64",
    requirements: Optional[str] = None,
    pipeline_file_paths: Optional[List[str]] = None,
):
    Path(output_dir).mkdir(parents=True, exist_ok=True)

    with open(os.path.join(output_dir, "Dockerfile"), "w") as docker_file:
        docker_file.writelines(
            [
                "FROM --platform=%s mysticai/pipeline-docker:latest" % platform,
            ]
        )

        if requirements is not None:
            docker_file.writelines(
                [
                    f"\nCOPY {requirements} /app/requirements.txt",
                    "\nRUN pip install -r requirements.txt",
                ]
            )
        docker_file.writelines(
            ["\nCOPY %s /app/pipelines/" % path for path in pipeline_graph_paths]
        )

        if pipeline_file_paths is not None:
            docker_file.writelines(
                ["\nCOPY %s /app/files/" % path for path in pipeline_file_paths]
            )


def create_docker_compose(path, gpu_index: Optional[str] = None, **environment_vars):
    presets_env_vars = {
        "LOG_LEVEL": "debug",
        "PYTHONDONTWRITEBYTECODE": 1,
        "PYTHONUNBUFFERED": 1,
        "DATABASE_HOSTNAME": "pipeline_db",
        "DATABASE_PORT": 5432,
        "DATABASE_USERNAME": "postgres",
        "DATABASE_PASSWORD": "example",
        "DATABASE_NAME": "pipeline",
        "DATABASE_CREATE_TABLES": 1,
        "REDIS_URL": "pipeline_redis",
    }

    for key in presets_env_vars:
        if key not in environment_vars:
            environment_vars[key] = presets_env_vars[key]

    docker_yml_dict = {
        "version": "3.9",
        "services": {
            "pipeline_api": {
                "image": "mysticai/pipeline-docker:latest",
                "container_name": "pipeline_api",
                "restart": "always",
                "ports": ["5010:80"],
                "build": {"context": "./", "dockerfile": "Dockerfile"},
                "environment": {
                    **environment_vars,
                },
                "depends_on": ["pipeline_db"],
            },
            "pipeline_db": {
                "image": "postgres",
                "container_name": "pipeline_db",
                "restart": "always",
                "environment": {"POSTGRES_PASSWORD": "example"},
            },
            "pipeline_redis": {
                "image": "bitnami/redis:latest",
                "container_name": "pipeline_redis",
                "restart": "always",
                "environment": ["ALLOW_EMPTY_PASSWORD=yes"],
                "ports": ["6379:6379"],
            },
        },
    }

    if gpu_index is not None:
        docker_yml_dict["services"]["pipeline_api"]["deploy"] = {
            "resources": {
                "reservations": {
                    "devices": [
                        {
                            "driver": "nvidia",
                            "device_ids": [f"{gpu_index}"],
                            "capabilities": ["gpu"],
                        }
                    ]
                }
            }
        }

    with open(os.path.join(path, "docker-compose.yml"), "w") as docker_compose_file:
        docker_compose_file.write(yaml.dump(docker_yml_dict))


def build():
    subprocess.run("sudo docker compose pull".split())
    subprocess.run("sudo docker compose build".split())



import os
import re
from datetime import datetime
from typing import Optional

import humps
from pydantic import BaseModel as PydanticModel
from pydantic import Extra, validator
from pydantic.generics import GenericModel as PydanticGenericModel

#: When set to True, model fields can be get/set by camelCase'd field names
#: This is useful for JS interop., where camelCase is the convention
#: (rather than Python's snake_case convention)
CAMEL_CASE_ALIASES = bool(os.environ.get("CAMEL_CASE_ALIASES", False))


def _generate_alias(s):
    return humps.camelize(s) if CAMEL_CASE_ALIASES else s


class BaseConfig:
    alias_generator = _generate_alias
    allow_population_by_field_name = True
    # Encode datetime objects to JSON as integer timestamps
    json_encoders = {datetime: datetime.timestamp}
    orm_mode = True


class BaseModel(PydanticModel):
    """Base model for schemas."""

    class Config(BaseConfig):
        pass


class GenericModel(PydanticGenericModel):
    """Base model for schemas with generic typing."""

    class Config(BaseConfig):
        pass


class Patchable(BaseModel):
    """Tag type for marking schemas for patching.

    A Patchable schema is one used for updating other schemas.
    """

    class Config:
        # Forbid fields not specified in the model; this prevents silently
        # failing updates to patch non-matching fields
        extra = Extra.forbid


class AvatarHolder(BaseModel):
    """Schema mixin for models which may hold avatars."""

    avatar_colour: Optional[str]
    avatar_image_url: Optional[str]

    @validator("avatar_colour")
    def validate_avatar_colour(cls, value: Optional[str]):
        if value is not None and not re.match(r"#[0-9a-f]{6}", value.lower()):
            raise ValueError("not a valid #rrggbb colour")
        return value



import enum
from typing import Optional

from .base import BaseModel


class FileFormat(str, enum.Enum):
    """Represents the different formats files can be uploaded in"""

    hex = "hex"
    binary = "binary"


class FileBase(BaseModel):
    name: str
    # hex is the default purely for backwards-compatability
    file_format: FileFormat = FileFormat.hex


class FileGet(FileBase):
    id: str
    path: str
    #: The data as hex-encoded bytes, if the data size is less than 20 kB
    data: Optional[str]
    #: The data size in bytes
    file_size: int


class FileCreate(FileBase):
    name: Optional[str]
    file_bytes: Optional[str]



class PipelineNotDeployed(Exception):
    def __init__(self, pipeline_id=None, message="Pipeline not deployed") -> None:
        self.pipeline_id = pipeline_id
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.pipeline_id} -> {self.message}"



from pipeline.objects.function import Function
from pipeline.objects.graph import Graph
from pipeline.objects.graph_node import GraphNode
from pipeline.objects.variable import Variable


class Pipeline:
    defined_pipelines = {}

    _current_pipeline: Graph
    _pipeline_context_active: bool = False
    _pipeline_context_name: str = None
    _compute_type: str = "gpu"
    _min_gpu_vram_mb: int = None

    def __init__(
        self,
        new_pipeline_name: str,
        compute_type: str = "gpu",
        min_gpu_vram_mb: int = None,
    ):
        self._pipeline_context_name = new_pipeline_name
        self._compute_type = compute_type
        self._min_gpu_vram_mb = min_gpu_vram_mb

    def __enter__(self):
        Pipeline._pipeline_context_active = True

        Pipeline._current_pipeline = Graph(
            name=self._pipeline_context_name,
            compute_type=self._compute_type,
            min_gpu_vram_mb=self._min_gpu_vram_mb,
        )

        return self

    def __exit__(self, type, value, traceback):
        Pipeline.defined_pipelines[
            Pipeline._current_pipeline.name
        ] = Pipeline._current_pipeline
        Pipeline._pipeline_context_active = False
        Pipeline._current_pipeline = None

    def output(self, *outputs: Variable) -> None:
        for _output in outputs:
            variable_index = Pipeline._current_pipeline.variables.index(_output)
            if variable_index != -1:
                Pipeline._current_pipeline.variables[variable_index].is_output = True

                Pipeline._current_pipeline.outputs.append(
                    Pipeline._current_pipeline.variables[variable_index]
                )
                for variable in Pipeline._current_pipeline.variables:
                    if variable.local_id == _output.local_id:
                        variable.is_output = True
                        break

    @staticmethod
    def get_pipeline(graph_name: str) -> Graph:
        """
        Get Graph representation of a pipeline.

            Parameters:
                    graph_name (str): name identifier of desired Pipeline

            Returns:
                    graph (Graph): Graph representation of Pipeline.
        """
        if graph_name in Pipeline.defined_pipelines:
            return Pipeline.defined_pipelines[graph_name]
        else:
            raise Exception("No Pipeline graph found with name '%s'" % graph_name)

    @staticmethod
    def add_variable(variable: Variable) -> None:
        if Pipeline._pipeline_context_active:

            if variable not in Pipeline._current_pipeline.variables:
                Pipeline._current_pipeline.variables.append(variable)
        else:
            raise Exception("Cant add a variable when not defining a pipeline!")

    @staticmethod
    def add_variables(*variables: Variable) -> None:
        for v in variables:
            Pipeline.add_variable(v)

    @staticmethod
    def add_function(function: Function) -> None:
        if Pipeline._pipeline_context_active:
            Pipeline._current_pipeline.functions.append(function)
        else:
            raise Exception("Cant add a function when not defining a pipeline!")

    @staticmethod
    def add_graph_node(graph_node: GraphNode) -> None:
        if Pipeline._pipeline_context_active:
            Pipeline._current_pipeline.nodes.append(graph_node)
        else:
            raise Exception("Cant add a node when not defining a pipeline!")



import argparse

from pipeline import PipelineCloud, configuration
from pipeline.util.logging import _print


def remote(args: argparse.Namespace) -> int:
    sub_command = getattr(args, "sub-command", None)

    if sub_command == "set":
        default_url = args.url
        configuration.config["DEFAULT_REMOTE"] = default_url
        configuration._save_config()
        _print(f"Set new default remote to '{configuration.config['DEFAULT_REMOTE']}'")
        return 0
    elif sub_command in ["list", "ls"]:
        remotes = [
            f"{_remote} (active)"
            if _remote == configuration.DEFAULT_REMOTE
            else f"{_remote}"
            for _remote in configuration.remote_auth.keys()
        ]
        _print("Authenticated remotes:")
        [print(_remote) for _remote in remotes]
        return 0
    elif sub_command == "login":
        valid_token = PipelineCloud._validate_token(args.token, args.url)

        if valid_token:
            configuration.remote_auth[args.url] = args.token
            configuration._save_auth()
            _print(f"Successfully authenticated with {args.url}")
            return 0

        _print(f"Couldn't authenticate with {args.url}", level="ERROR")
        return 1



import inspect
import uuid
from hashlib import sha256
from typing import Any

from pipeline.schemas.model import ModelGet
from pipeline.util import generate_id, hex_to_python_object


class Model:
    local_id: str
    remote_id: str

    name: str
    source: str
    hash: str

    model: Any

    def __init__(self, model: Any, *, name: str = "", local_id: str = None):

        self.name = name
        self.model = model
        try:
            self.source = inspect.getsource(model.__class__)
        except OSError:
            self.source = str(uuid.uuid4())

        self.hash = sha256(self.source.encode()).hexdigest()
        self.local_id = generate_id(10) if local_id is None else local_id
        if not hasattr(self.model, "local_id"):
            setattr(self.model, "local_id", self.local_id)

    @classmethod
    def from_schema(cls, schema: ModelGet):
        pickled_data = hex_to_python_object(schema.hex_file.data)
        if isinstance(pickled_data, Model):
            pickled_data.local_id = schema.id
            return pickled_data
        return cls(
            pickled_data,
            name=schema.name,
            local_id=schema.id,
        )



from typing import List

from .base import BaseModel
from .resource import ResourceCreate, ResourceGet


class WorkerBase(BaseModel):
    worker_ip: str
    worker_name: str


class WorkerCreate(WorkerBase):
    resources: List[ResourceCreate]


class WorkerGet(WorkerBase):
    id: str
    resources: List[ResourceGet]



from uuid import uuid4

from pydantic import BaseModel, Field


class RedisCommandSchema(BaseModel):
    command_thread_id: str = Field(default_factory=lambda: str(uuid4()))
    command: str
    data: dict
    to: str
    sender: str



import io
import random
import string
from typing import Any, Callable, Optional

from cloudpickle import dumps
from dill import loads

from pipeline.schemas.file import FileCreate


def generate_id(length: int) -> str:
    return "".join((random.choice(string.ascii_letters) for i in range(length)))


def python_object_to_hex(obj: Any) -> str:
    return dumps(obj).hex()


def hex_to_python_object(hex: str) -> Any:
    h = bytes.fromhex(hex)
    return loads(h)


def python_object_to_name(obj: Any) -> Optional[str]:
    try:
        name = obj.__name__
    except Exception:
        name = None
    return name


def python_object_to_file_create(obj: Any, name: str = None):
    if name is None:
        name = generate_id(20)

    return FileCreate(name=name, file_bytes=python_object_to_hex)


class CallbackBytesIO(io.BytesIO):
    """Provides same interface as BytesIO but additionally calls a callback function
    whenever the 'read' method is called.

    This is similar to tqdm's own CallbackIOWrapper but this does not play nicely with
    all features of httpx so we use our own in some cases.
    """

    def __init__(self, callback: Callable, initial_bytes: bytes):
        self._callback = callback
        super().__init__(initial_bytes)

    def read(self, size=-1) -> bytes:
        data = super().read(size)
        self._callback(len(data))
        return data



from functools import partial, wraps

from pipeline.objects.function import Function
from pipeline.objects.graph_node import GraphNode
from pipeline.objects.model import Model
from pipeline.objects.pipeline import Pipeline
from pipeline.objects.variable import Variable


def pipeline_function(function=None, *, run_once=False, on_startup=False):
    """_summary_

    Args:
        function (callable, optional): _description_. This is the function to be
            wrapped, you do not pass this in manually it's automatically handled.

        run_once (bool, optional): _description_. Defaults to False. Setting to True
            will ensure that the decorated funciton is only called once
            in a pipeline run

        on_startup (bool, optional): _description_. Defaults to False. Setting to True
            will cause the wrapped function to be executed at the start of a pipeline
            run, regardless of when it's placed when defining the pipeline.

    """
    if function is None:
        return partial(pipeline_function, run_once=run_once, on_startup=on_startup)

    @wraps(function)
    def execute_func(*args, **kwargs):
        if not Pipeline._pipeline_context_active:
            return function(*args, **kwargs)
        else:
            function_ios = function.__annotations__
            if "return" not in function_ios:
                raise Exception(
                    "Must include an output type e.g. 'def my_func(...) -> int:'"
                )

            processed_args: Variable = []
            for input_arg in args:
                if isinstance(input_arg, Variable):
                    processed_args.append(input_arg)
                elif hasattr(input_arg, "__pipeline_model__"):
                    if function.__pipeline_function__.class_instance is None:
                        function.__pipeline_function__.class_instance = input_arg
                else:
                    raise Exception(
                        (
                            "You can't input random variables, "
                            "follow the way of the Pipeline. Got type %s"
                            % type(input_arg)
                        )
                    )

            node_output = Variable(type_class=function.__annotations__["return"])
            Pipeline.add_variable(node_output)
            Pipeline.add_function(function.__pipeline_function__)
            new_node = GraphNode(
                function=function.__pipeline_function__,
                inputs=processed_args,
                outputs=[node_output],
            )
            Pipeline.add_graph_node(new_node)

            return node_output

    execute_func.__function__ = function

    function.__run_once__ = run_once
    function.__has_run__ = False

    function.__on_startup__ = on_startup
    function.__pipeline_function__ = Function(function)

    return execute_func


class pipeline_model(object):
    def __init__(
        self,
        model_class=None,
    ):
        if model_class is not None:
            model_class.__pipeline_model__ = True

        self.model_class = model_class

    def __call__(self, *args, **kwargs):

        if len(args) + len(kwargs) == 1:
            self.model_class = args[0]
            self.model_class.__pipeline_model__ = True
            return self.__function_exe__
        else:
            return self.__function_exe__(*args, **kwargs)

    def __function_exe__(self, *args, **kwargs):
        if not Pipeline._current_pipeline:
            return self.model_class(*args, **kwargs)
        else:
            created_model = self.model_class(*args, **kwargs)

            model_schema = Model(model=created_model)
            Pipeline._current_pipeline.models.append(model_schema)
            return created_model



from pipeline.objects import (
    Graph,
    Pipeline,
    PipelineFile,
    Variable,
    pipeline_function,
    pipeline_model,
)


def onnx_to_pipeline(path: str, name: str = "ONNX model") -> Graph:
    """
    Create a pipeline from an onnx model file
        Parameters:
                path (str): local path to onnx model file
                name (str): Desired name to be given to this pipeline

        Returns:
                pipeline (Graph): Executable Pipeline Graph object
    """

    @pipeline_model
    class model:
        def __init__(self):
            self.session = None

        import numpy as np

        @pipeline_function
        def predict(self, onnx_output: list, onnx_input: dict = {}) -> list:
            res = self.session.run(onnx_output, onnx_input)
            return res[0].tolist()

        @pipeline_function(run_once=True, on_startup=True)
        def load(self, onnx_file: PipelineFile) -> bool:
            import onnxruntime

            self.session = onnxruntime.InferenceSession(
                onnx_file.path,
                providers=[
                    "CUDAExecutionProvider",
                ],
            )
            return True

    with Pipeline(name) as pipeline:
        onnx_file = PipelineFile(path=path)
        onnx_output = Variable(list, is_input=True)
        onnx_input = Variable(dict, is_input=True)

        pipeline.add_variables(
            onnx_file,
            onnx_output,
            onnx_input,
        )

        model = model()
        model.load(onnx_file)

        output = model.predict(
            onnx_output,
            onnx_input,
        )

        pipeline.output(output)

    return Pipeline.get_pipeline(name)



from typing import List

from cloudpickle import dumps
from dill import loads

from pipeline.objects.function import Function
from pipeline.objects.graph_node import GraphNode
from pipeline.objects.model import Model
from pipeline.objects.variable import PipelineFile, Variable
from pipeline.schemas.pipeline import PipelineGet
from pipeline.util import generate_id


class Graph:
    local_id: str
    remote_id: str

    name: str

    functions: List[Function]
    variables: List[Variable]

    outputs: List[Variable]

    nodes: List[GraphNode]

    models: List[Model]

    # TODO: Add generic objects (e.g. Model) to be included in the graph

    compute_type: str
    min_gpu_vram_mb: int

    def __init__(
        self,
        *,
        name: str = "",
        variables: List[Variable] = None,
        functions: List[Function] = None,
        outputs: List[Variable] = None,
        nodes: List[GraphNode] = None,
        models: List[Model] = None,
        compute_type: str = "gpu",
        min_gpu_vram_mb: int = None,
    ):
        self.name = name
        self.local_id = generate_id(10)

        self.variables = variables if variables is not None else []
        self.functions = functions if functions is not None else []
        self.outputs = outputs if outputs is not None else []
        self.nodes = nodes if nodes is not None else []
        self.models = models if models is not None else []
        # Flag set when all functions with the on_startup field have run
        self._has_run_startup = False
        self.compute_type = compute_type
        self.min_gpu_vram_mb = min_gpu_vram_mb

    def _startup(self):
        if self._has_run_startup:
            return

        startup_variables = {}

        for var in self.variables:
            # At the moment only the PipelineFile variable can be used on startup
            if isinstance(var, PipelineFile):
                startup_variables[var.local_id] = var

        for node in self.nodes:

            node_inputs: List[Variable] = []
            node_function: Function = None

            for function in self.functions:
                if function.local_id == node.function.local_id:
                    node_function = function
                    break
            if (
                getattr(node_function.function, "__run_once__", False)
                and not getattr(node_function.function, "__has_run__", False)
            ) or not getattr(node_function.function, "__on_startup__", False):
                continue

            for _node_input in node.inputs:
                for variable in self.variables:
                    if variable.local_id == _node_input.local_id:
                        node_inputs.append(variable)
                        break

            function_inputs = []
            for _input in node_inputs:
                function_inputs.append(startup_variables[_input.local_id])

            if node_function.function is None:
                raise Exception(
                    "Node function is None (id:%s)" % node.function.local_id
                )

            if getattr(node_function, "class_instance", None) is not None:
                node_function.function(node_function.class_instance, *function_inputs)
            else:
                node_function.function(*function_inputs)

            if getattr(node_function.function, "__has_run__", False):
                node_function.function.__has_run__ = True

        self._has_run_startup = True

    def run(self, *inputs):
        input_variables: List[Variable] = [
            var for var in self.variables if var.is_input
        ]

        if len(inputs) != len(input_variables):
            raise Exception(
                "Mismatch of number of inputs, expecting %u got %s"
                % (len(input_variables), len(inputs))
            )

        self._startup()

        running_variables = {}

        # Add all PipelineFile's to the running variables
        for var in self.variables:
            if isinstance(var, PipelineFile):
                if var.remote_id is not None and not var.path:
                    raise Exception(
                        "Must call PipelineCloud().download_remotes(...) on "
                        "remote PipelineFiles"
                    )

                running_variables[var.local_id] = var

        for i, input in enumerate(inputs):
            if not isinstance(input, input_variables[i].type_class):
                raise Exception(
                    "Input type mismatch, expceted %s got %s"
                    % (
                        input_variables[i].type_class,
                        input.__class__,
                    )
                )
            running_variables[input_variables[i].local_id] = input

        for node in self.nodes:

            node_inputs: List[Variable] = []
            node_outputs: List[Variable] = []
            node_function: Function = None

            for function in self.functions:
                if function.local_id == node.function.local_id:
                    node_function = function
                    break

            if getattr(node_function.function, "__run_once__", False) and getattr(
                node_function.function, "__has_run__", False
            ):
                continue

            for _node_input in node.inputs:
                for variable in self.variables:
                    if variable.local_id == _node_input.local_id:
                        node_inputs.append(variable)
                        break
            for _node_output in node.outputs:
                for variable in self.variables:
                    if variable.local_id == _node_output.local_id:
                        node_outputs.append(variable)
                        break

            function_inputs = []
            for _input in node_inputs:
                function_inputs.append(running_variables[_input.local_id])

            if node_function.function is None:
                raise Exception(
                    "Node function is none (id:%s)" % node.function.local_id
                )

            if getattr(node_function, "class_instance", None) is not None:
                output = node_function.function(
                    node_function.class_instance, *function_inputs
                )
            else:
                output = node_function.function(*function_inputs)

            running_variables[node_outputs[0].local_id] = output

            if not getattr(node_function.function, "__has_run__", False):
                node_function.function.__has_run__ = True

        return_variables = []

        for output_variable in self.outputs:
            return_variables.append(running_variables[output_variable.local_id])

        return return_variables

    def _update_function_local_id(self, old_id: str, new_id: str) -> None:
        for func in self.functions:
            if func.local_id == old_id:
                func.local_id = new_id
                return
        raise Exception("Function with local_id:%s not found" % old_id)

    @classmethod
    def from_schema(cls, schema: PipelineGet):
        variables = [Variable.from_schema(_var) for _var in schema.variables]
        functions = [Function.from_schema(_func) for _func in schema.functions]
        models = [Model.from_schema(_model) for _model in schema.models]

        # Rebind functions -> models
        update_functions = []
        for _func in functions:
            if hasattr(_func.class_instance, "__pipeline_model__"):
                model = _func.class_instance
                is_bound = False
                for _model in models:
                    if _model.model.local_id == model.local_id:
                        bound_method = _func.function.__get__(
                            _model.model, _model.model.__class__
                        )
                        setattr(_model.model, _func.function.__name__, bound_method)
                        is_bound = True
                        _func.class_instance = _model.model
                if not is_bound:
                    raise Exception(
                        "Did not find a class to bind for model (local_id:%s)"
                        % model.local_id
                    )
            elif _func.class_instance is not None:
                raise Exception(
                    "Incorrect bound class:%s\ndir:%s"
                    % (_func.class_instance, dir(_func.class_instance))
                )
            update_functions.append(_func)
        functions = update_functions

        outputs = []
        for _output in schema.outputs:
            for _var in variables:
                if _var.local_id == _output:
                    outputs.append(_var)

        nodes = []

        for _node in schema.graph_nodes:
            node_inputs = []
            for node_str in _node.inputs:
                for _var in variables:
                    if _var.local_id == node_str:
                        node_inputs.append(_var)
                        break

            node_outputs = []
            for node_str in _node.outputs:
                for _var in variables:
                    if _var.local_id == node_str:
                        node_outputs.append(_var)
                        break

            function = None

            for _func in functions:
                if _func.local_id == _node.function:
                    function = _func
                    break

            if function is None:
                raise Exception("Function not found:%s" % _node.function)

            nodes.append(
                GraphNode(
                    function=function,
                    inputs=node_inputs,
                    outputs=node_outputs,
                    local_id=_node.local_id,
                )
            )

        remade_graph = cls(
            name=schema.name,
            variables=variables,
            functions=functions,
            outputs=outputs,
            nodes=nodes,
            models=models,
        )

        return remade_graph

    def save(self, save_path):
        with open(save_path, "wb") as save_file:
            save_file.write(dumps(self))

    @classmethod
    def load(cls, load_path):
        with open(load_path, "rb") as load_file:
            return loads(load_file.read())



from __future__ import annotations

import io
import json
import urllib.parse
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Type, Union

import httpx

from pipeline.exceptions.InvalidSchema import InvalidSchema
from pipeline.schemas.base import BaseModel
from pipeline.schemas.compute_requirements import ComputeRequirements
from pipeline.schemas.data import DataGet
from pipeline.schemas.file import FileCreate, FileGet
from pipeline.schemas.function import FunctionGet
from pipeline.schemas.model import ModelGet
from pipeline.schemas.pipeline import PipelineFileVariableGet, PipelineGet
from pipeline.schemas.pipeline_file import MultipartUploadMetadata, PipelineFileGet
from pipeline.schemas.run import RunCreate, RunGet
from pipeline.util import generate_id, python_object_to_hex

if TYPE_CHECKING:
    from pipeline.objects import Function, Graph, Model

from pipeline.api import PipelineCloud as _SyncPipelineCloud

FILE_CHUNK_SIZE = 200 * 1024 * 1024  # 200 MiB


class PipelineCloud(_SyncPipelineCloud):
    def _raise_not_implemeneted(self):
        raise NotImplementedError("This function is not implemented")

    def upload_python_object_to_file(self, obj) -> FileGet:
        self._raise_not_implemeneted()

    def _initialise_direct_pipeline_file_upload(self, file_size: int) -> str:
        self._raise_not_implemeneted()

    def _direct_upload_pipeline_file_chunk(
        self, data: bytes, pipeline_file_id: str, part_num: int
    ) -> MultipartUploadMetadata:
        self._raise_not_implemeneted()

    def _finalise_direct_pipeline_file_upload(
        self, pipeline_file_id: str, multipart_metadata: List[MultipartUploadMetadata]
    ) -> PipelineFileGet:
        self._raise_not_implemeneted()

    def upload_pipeline_file(self, pipeline_file) -> PipelineFileVariableGet:
        self._raise_not_implemeneted()

    def _get(self, endpoint: str, params: Dict[str, Any] = None):
        self._raise_not_implemeneted()

    def upload_function(self, function: Function) -> FunctionGet:
        self._raise_not_implemeneted()

    def upload_model(self, model: Model) -> ModelGet:
        self._raise_not_implemeneted()

    def upload_pipeline(
        self,
        new_pipeline_graph: Graph,
        public: bool = False,
        description: str = "",
        tags: Set[str] = None,
    ) -> PipelineGet:
        self._raise_not_implemeneted()

    def _download_schema(
        self, schema: Type[BaseModel], endpoint: str, params: Optional[Dict[str, Any]]
    ) -> Type[BaseModel]:
        self._raise_not_implemeneted()

    def download_function(self, id: str) -> Function:
        self._raise_not_implemeneted()

    def download_model(self, id: str) -> Model:
        self._raise_not_implemeneted()

    def download_data(self, id: str) -> Any:
        self._raise_not_implemeneted()

    def download_result(self, result_id_or_schema: Union[str, RunGet]) -> Any:
        self._raise_not_implemeneted()

    def download_pipeline(self, id: str) -> Graph:
        self._raise_not_implemeneted()

    async def _post_file(self, endpoint: str, file: io.BytesIO) -> FileGet:
        if not hasattr(file, "name"):
            file.name = generate_id(20)

        headers = {
            "Authorization": "Bearer %s" % self.token,
        }
        url = urllib.parse.urljoin(self.url, endpoint)
        async with httpx.AsyncClient() as client:
            response = await client.post(
                url,
                headers=headers,
                files={"file": (file.name, file, "application/octet-stream")},
                timeout=self.timeout,
            )
        if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
            schema = FileCreate.__name__
            raise InvalidSchema(schema=schema)
        else:
            response.raise_for_status()
        return FileGet.parse_obj(response.json())

    async def upload_file(self, file_or_path) -> FileGet:

        if isinstance(file_or_path, str):
            # TODO: Change this to wrap the file object reader to convert to hex
            # everytime anything is read instead of reading it all at once.

            with open(file_or_path, "rb") as file:
                buffer = file.read()
            hex_buffer = buffer.hex()
            return await self._post_file("/v2/files/", io.BytesIO(hex_buffer.encode()))
        else:
            return await self._post_file("/v2/files/", file_or_path)

    async def upload_data(self, file_or_path) -> DataGet:
        uploaded_file = await self.upload_file(file_or_path)
        uploaded_data = await self._post("/v2/data", uploaded_file.dict())
        return DataGet.parse_obj(uploaded_data)

    async def _post(self, endpoint, json_data):
        headers = {
            "Authorization": "Bearer %s" % self.token,
            "Content-type": "application/json",
        }

        url = urllib.parse.urljoin(self.url, endpoint)
        async with httpx.AsyncClient() as client:
            response = await client.post(
                url, headers=headers, json=json_data, timeout=self.timeout
            )

        if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
            schema = json_data
            raise InvalidSchema(schema=schema)
        else:
            self._get_raise_for_status(response)

        return response.json()

    async def run_pipeline(
        self,
        pipeline_id_or_schema: Union[str, PipelineGet],
        raw_data_or_schema: Union[Any, DataGet],
        compute_type: Optional[str] = None,
        min_gpu_vram_mb: Optional[int] = None,
    ):
        """
        Uploads Data and executes a Run of given pipeline over given data.

            Parameters:
                    pipeline_id_or_schema (Union[str, PipelineGet]):
                        The id for the desired pipeline
                        or the schema obtained from uploading it
                    raw_data_or_schema (Union[Any, DataGet]):
                        Raw data for Pipeline execution
                        or schema for already uploaded data.
                    compute_type (Optional[str]):
                        Compute type requirement for this run (e.g. "cpu" or "gpu"). If
                        not defined, it will fallback to that specified by the Pipeline
                        itself (typically this is "gpu")
                    min_gpu_vram_mb (Optional[int]):
                        Minimum amount of GPU VRAM required. If not defined, it will
                        fallback to that specified by the Pipeline (if applicable). This
                        should only be specified for GPU workloads.

            Returns:
                    run (Any): Run object containing metadata and outputs.
        """
        # TODO: Add support for generic object inference. Only strs at the moment.
        if not isinstance(raw_data_or_schema, DataGet):
            temp_file = io.BytesIO(python_object_to_hex(raw_data_or_schema).encode())
            uploaded_data = await self.upload_data(temp_file)
            _data_id = uploaded_data.id
        elif isinstance(raw_data_or_schema, DataGet):
            _data_id = raw_data_or_schema.id
        else:
            raise Exception("Must either pass a raw data, or DataGet schema.")

        pipeline_id = None
        if isinstance(pipeline_id_or_schema, str):
            pipeline_id = pipeline_id_or_schema
        elif isinstance(pipeline_id_or_schema, PipelineGet):
            pipeline_id = pipeline_id_or_schema.id
        else:
            raise InvalidSchema(
                schema=pipeline_id_or_schema,
                message=(
                    "Must either pass a pipeline id, or a pipeline get schema. "
                    "Not object of type %s in arg 1." % str(pipeline_id_or_schema)
                ),
            )

        compute_requirements = None
        if min_gpu_vram_mb:
            compute_requirements = ComputeRequirements(min_gpu_vram_mb=min_gpu_vram_mb)

        run_create_schema = RunCreate(
            pipeline_id=pipeline_id,
            data_id=_data_id,
            compute_type=compute_type,
            compute_requirements=compute_requirements,
        )
        run_json: dict = await self._post(
            "/v2/runs", json.loads(run_create_schema.json())
        )
        return RunGet.parse_obj(run_json)



from pipeline import configuration
from pipeline.api.cloud import PipelineCloud
from pipeline.objects import (
    Pipeline,
    PipelineFile,
    Variable,
    onnx_to_pipeline,
    pipeline_function,
    pipeline_model,
)

__all__ = [
    "Pipeline",
    "Variable",
    "pipeline_model",
    "pipeline_function",
    "PipelineCloud",
    "PipelineFile",
    "onnx_to_pipeline",
    "configuration",
]



from datetime import datetime
from typing import Optional

from pydantic import validator

from .base import BaseModel, Patchable


class TokenGet(BaseModel):
    """API token representation when returned from an API call."""

    id: str
    #: Token value, used when authenticating with the API
    value: str
    #: Token value, used when authenticating with the API
    name: str
    #: Timestamp value of creation
    created_at: datetime
    #: Timestamp value of last update
    updated_at: datetime
    #: Timestamp value of expiry
    expires_at: Optional[datetime]
    #: Timestamp value of last usage
    last_used: Optional[datetime]
    #: If token is active (set by user and it's before its expiry date)
    is_active: bool
    #: Arbitrarily set user flag for token validity
    is_enabled: bool
    #: User associated with this token (if user token)
    user_id: Optional[str]


class TokenCreate(BaseModel):
    """Model for creating token"""

    #: Arbitrary name to be assigned to token
    name: Optional[str]
    #: Role name which assigns permissions
    type: str
    #: Timestamp value of future expiry
    expires_at: Optional[datetime]


class TokenPatch(Patchable):
    """Model for patching token"""

    # Arbitrary name to be assigned to token
    name: Optional[str]
    # Boolean specifying if token can be used
    is_enabled: Optional[bool]
    # Token value (only sudo tokens can use this)
    value: Optional[str]
    # Datetime to expire the token
    expire_at: Optional[datetime]

    @validator("is_enabled")
    def prevent_none(cls, v):
        if v is None:
            raise ValueError("must not be null")
        return v



import enum
from datetime import datetime
from typing import Optional

from .base import BaseModel
from .file import FileGet
from .token import TokenGet


class FileType(enum.Enum):
    text = "text"
    image = "image"


class DataGet(BaseModel):
    id: str
    hex_file: FileGet
    created_at: datetime
    modified_at: Optional[datetime]
    name: Optional[str]
    size: Optional[int]
    file_type: Optional[FileType]
    token_created_by: Optional[TokenGet]
    token_modified_by: Optional[TokenGet]
    url: Optional[str]
    preview: Optional[str]



from typing import List, Optional

from pydantic import Field, root_validator

from pipeline.schemas.base import BaseModel
from pipeline.schemas.compute_requirements import ComputeRequirements, ComputeType
from pipeline.schemas.file import FileCreate, FileGet
from pipeline.schemas.runnable import RunnableGet, RunnableType


class FunctionBase(BaseModel):
    id: Optional[str]
    name: str


class FunctionGet(RunnableGet):
    id: str
    hex_file: FileGet

    source_sample: str
    type: RunnableType = Field(RunnableType.function, const=True)

    class Config:
        orm_mode = True


class FunctionIO(BaseModel):
    """Descriptive schema of a function's input/output.

    Not intended to be sufficient to create an input/output type object.
    """

    # If present, the name of the argument
    name: str
    # The name of the type, if available (some types cannot have their names
    # extracted)
    type_name: str


class FunctionGetDetailed(FunctionGet):
    inputs: List[FunctionIO]
    output: List[FunctionIO]


class FunctionIOCreate(BaseModel):
    name: str
    file_id: Optional[str]
    file: Optional[FileCreate]

    @root_validator
    def file_or_id_validation(cls, values):
        file, file_id = values.get("file"), values.get("file_id")

        file_defined = file is not None
        file_id_defined = file_id is not None

        if file_defined == file_id_defined:
            raise ValueError(
                "You must define either the file OR file_id of a function."
            )

        return values


class FunctionCreate(BaseModel):
    # The local ID is assigned when a new function is used as part of a new
    # pipeline; the server uses the local ID to associated a function to a
    # graph node before replacing the local ID with the server-generated one
    local_id: Optional[str]

    # function_hex: str
    function_source: str

    inputs: List[FunctionIO]
    output: List[FunctionIO]

    name: str
    hash: str

    file_id: Optional[str]
    file: Optional[FileCreate]

    # By default a Function will require GPU resources
    compute_type: ComputeType = ComputeType.gpu
    compute_requirements: Optional[ComputeRequirements]

    @root_validator
    def file_or_id_validation(cls, values):
        file, file_id = values.get("file"), values.get("file_id")

        file_defined = file is not None
        file_id_defined = file_id is not None

        if file_defined == file_id_defined:
            raise ValueError(
                "You must define either the file OR file_id of a function."
            )

        return values



import enum
from typing import List, Optional

from pydantic import Field

from .base import BaseModel


class RunnableType(enum.Enum):
    function = "function"
    pipeline = "pipeline"


class RunnableIOGet(BaseModel):
    """Unrealised/expected input/output data to a Runnable."""

    name: str
    type: str


class RunnableGet(BaseModel):
    id: str
    type: str
    name: Optional[str]


class RunnableGetDetailed(RunnableGet):
    expected_inputs: List[RunnableIOGet] = []
    expected_outputs: List[RunnableIOGet] = []
    code_excerpt: Optional[str] = None
    last_runs = []


# NOTE QUESTION: do we use these classes?
class FunctionGet(RunnableGet):
    type: RunnableType = Field(RunnableType.function, const=True)


class FunctionGetDetailed(FunctionGet, RunnableGetDetailed):
    pass


class PipelineGet(RunnableGet):
    type: RunnableType = Field(RunnableType.pipeline, const=True)


class PipelineGetDetailed(PipelineGet, RunnableGetDetailed):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pipeline',
 'pipeline.api',
 'pipeline.api.asyncio',
 'pipeline.console',
 'pipeline.docker',
 'pipeline.exceptions',
 'pipeline.objects',
 'pipeline.objects.environment',
 'pipeline.objects.huggingface',
 'pipeline.schemas',
 'pipeline.schemas.redis',
 'pipeline.util',
 'pipeline.util.torch_utils']

package_data = \
{'': ['*']}

install_requires = \
['cloudpickle>=2.2.0,<3.0.0',
 'dill>=0.3.6,<0.4.0',
 'httpx>=0.23.1,<0.24.0',
 'packaging>=21.0',
 'pydantic>=1.8.2,<2.0.0',
 'pyhumps>=3.0.2,<4.0.0',
 'setuptools>=65.4.1,<66.0.0',
 'tabulate>=0.9.0,<0.10.0',
 'tqdm>=4.62.3,<5.0.0']

extras_require = \
{'docker': ['PyYAML>=6.0,<7.0']}

entry_points = \
{'console_scripts': ['pipeline = pipeline.console:_run']}

setup_kwargs = {
    'name': 'pipeline-ai',
    'version': '0.3.5',
    'description': 'Pipelines for machine learning workloads.',
    'long_description': '# [Pipeline](https://pipeline.ai) [![Production workflow](https://github.com/neuro-ai-dev/pipeline/actions/workflows/prod-wf.yml/badge.svg?branch=main)](https://github.com/neuro-ai-dev/pipeline/actions/workflows/prod-wf.yml) [![Version](https://img.shields.io/pypi/v/pipeline-ai)](https://pypi.org/project/pipeline-ai) ![Size](https://img.shields.io/github/repo-size/neuro-ai-dev/pipeline) ![Downloads](https://img.shields.io/pypi/dm/pipeline-ai) [![License](https://img.shields.io/crates/l/ap)](https://www.apache.org/licenses/LICENSE-2.0) [![Discord](https://img.shields.io/badge/discord-join-blue)](https://discord.gg/eJQRkBdEcs)\n\n[_powered by mystic_](https://www.mystic.ai/)\n\n# Table of Contents\n\n- [About](#about)\n- [Version roadmap](#version-roadmap)\n  - [v0.4.0](#v040-jan-2023)\n  - [v0.5.0](#v050-janfeb-2023)\n  - [No current specified version features](#no-current-specified-version-features)\n- [Quickstart](#quickstart)\n  - [Basic maths](#basic-maths)\n  - [Transformers (GPT-Neo 125M)](#transformers-gpt-neo-125m)\n- [Installation instructions](#installation-instructions)\n  - [Linux, Mac (intel)](#linux--mac--intel-)\n  - [Mac (arm/M1)](#mac--arm-m1-)\n- [Development](#development)\n- [License](#license)\n\n# About\n\nPipeline is a python library that provides a simple way to construct computational graphs for AI/ML. The library is suitable for both development and production environments supporting inference and training/finetuning. This library is also a direct interface to [Pipeline.ai](https://pipeline.ai) which provides a compute engine to run pipelines at scale and on enterprise GPUs.\n\nThe syntax used for defining AI/ML pipelines shares some similarities in syntax to sessions in [Tensorflow v1](https://www.tensorflow.org/api_docs/python/tf/compat/v1/InteractiveSession), and Flows found in [Prefect](https://github.com/PrefectHQ/prefect). In future releases we will be moving away from this syntax to a C based graph compiler which interprets python directly (and other languages) allowing users of the API to compose graphs in a more native way to the chosen language.\n# Version roadmap\n\n## v0.4.0 (Jan 2023)\n\n- Custom environments on PipelineCloud (remote compute services)\n- Kwarg inputs to runs\n- Extended IO inputs to `pipeline_function` objects\n\n## v0.5.0 (Jan/Feb 2023)\n\n- Pipeline chaining\n- `if` statements & `while/for` loops\n\n## No current specified version features\n\n- Run log streaming\n- Run progress tracking\n- Resource dedication\n- Pipeline scecific remote load balancer (10% of traffic to one pipeline 80% to another)\n- Usage capping\n- Run result streaming\n- Progromatic autoscaling\n- Alerts\n- Events\n- Different python versions on remote compute services\n# Quickstart\n\n> :warning: **Uploading pipelines to Pipeline Cloud works best in Python 3.9.** We strongly recommend you use Python 3.9 when uploading pipelines because the `pipeline-ai` library is still in beta and is known to cause opaque errors when pipelines are serialised from a non-3.9 environment.\n\n\n## Basic maths\n\n```python\nfrom pipeline import Pipeline, Variable, pipeline_function\n\n\n@pipeline_function\ndef square(a: float) -> float:\n    return a**2\n\n@pipeline_function\ndef multiply(a: float, b: float) -> float:\n    return a * b\n\nwith Pipeline("maths") as pipeline:\n    flt_1 = Variable(type_class=float, is_input=True)\n    flt_2 = Variable(type_class=float, is_input=True)\n    pipeline.add_variables(flt_1, flt_2)\n\n    sq_1 = square(flt_1)\n    res_1 = multiply(flt_2, sq_1)\n    pipeline.output(res_!)\n\noutput_pipeline = Pipeline.get_pipeline("maths")\nprint(output_pipeline.run(5.0, 6.0))\n\n```\n\n## Transformers (GPT-Neo 125M)\n\n_Note: requires `torch` and `transformers` as dependencies._\n```python\nfrom pipeline import Pipeline, Variable\nfrom pipeline.objects.huggingface.TransformersModelForCausalLM import (\n    TransformersModelForCausalLM,\n)\n\nwith Pipeline("HF pipeline") as builder:\n    input_str = Variable(str, is_input=True)\n    model_kwargs = Variable(dict, is_input=True)\n\n    builder.add_variables(input_str, model_kwargs)\n\n    hf_model = TransformersModelForCausalLM(\n        model_path="EleutherAI/gpt-neo-125M",\n        tokenizer_path="EleutherAI/gpt-neo-125M",\n    )\n    hf_model.load()\n    output_str = hf_model.predict(input_str, model_kwargs)\n\n    builder.output(output_str)\n\noutput_pipeline = Pipeline.get_pipeline("HF pipeline")\n\nprint(\n    output_pipeline.run(\n        "Hello my name is", {"min_length": 100, "max_length": 150, "temperature": 0.5}\n    )\n)\n```\n\n# Installation instructions\n\n## Linux, Mac (intel)\n\n```shell\npip install -U pipeline-ai\n```\n\n## Mac (arm/M1)\n\nDue to the ARM architecture of the M1 core it is necessary to take additional steps to install Pipeline, mostly due to the transformers library. We recoomend running inside of a conda environment as shown below.\n\n1. Make sure Rosetta2 is disabled.\n2. From terminal run:\n\n```\nxcode-select --install\n```\n\n3. Install Miniforge, instructions here: [https://github.com/conda-forge/miniforge](https://github.com/conda-forge/miniforge) or follow the below:\n   1. Download the Miniforge install script here: [https://github.com/conda-forge/miniforge/releases/latest/download/Miniforge3-MacOSX-arm64.sh](https://github.com/conda-forge/miniforge/releases/latest/download/Miniforge3-MacOSX-arm64.sh)\n   2. Make the shell executable and run\n   ```\n   sudo chmod 775 Miniforge3-MacOSX-arm64.sh\n   ./Miniforge3-MacOSX-arm64.sh\n   ```\n4. Create a conda based virtual env and activate:\n\n```\nconda create --name pipeline-env python=3.9\nconda activate pipeline-env\n```\n\n5. Install tensorflow\n\n```\nconda install -c apple tensorflow-deps\npython -m pip install -U pip\npython -m pip install -U tensorflow-macos\npython -m pip install -U tensorflow-metal\n```\n\n6. Install transformers\n\n```\nconda install -c huggingface transformers -y\n```\n\n7. Install pipeline\n\n```\npython -m pip install -U pipeline-ai\n```\n\n\n\n# Development\n\nThis project is made with poetry, [so firstly setup poetry on your machine](https://python-poetry.org/docs/#installation).\n\nOnce that is done, please run\n\n    sh setup.sh\n\nWith this you should be good to go. This sets up dependencies, pre-commit hooks and\npre-push hooks.\n\nYou can manually run pre commit hooks with\n\n    pre-commit run --all-files\n\nTo run tests manually please run\n\n    pytest\n\n# License\n\nPipeline is licensed under [Apache Software License Version 2.0](https://www.apache.org/licenses/LICENSE-2.0).\n',
    'author': 'Paul Hetherington',
    'author_email': 'ph@mystic.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import re

# TO-DO: refactor validators as below
# https://pydantic-docs.helpmanual.io/usage/validators/

# Email regex mostly following RFC2822 specification. Covers ~99% of emails in use today
# Allows groups of alphanumerics and some special characters separated by dots,
# followed by a @,
# followed by groups of alphanumerics and non-staring/non-ending dashes,
# separated by dots.
EMAIL_REGEX = re.compile(
    r"[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*"
    r"@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)"
    r"+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?"
)
# Simple password regex, requires a minimum of 8 characters with at least one
# uppercase letter, one lowercase letter, and one number.
PASSWORD_REGEX = re.compile(r"^(?=.*\d)(?=.*[a-z])(?=.*[A-Z]).{8,}$")
# Simple username regex, number of characters between 3-24, allowing only alphanumerics,
# dashes and underscores.
USERNAME_REGEX = re.compile(r"^[a-zA-Z0-9-_]{3,24}$")


def valid_email(email_string: str) -> bool:
    return EMAIL_REGEX.match(email_string) is not None


def valid_password(password_string: str) -> bool:
    return PASSWORD_REGEX.match(password_string) is not None


def valid_username(username_string: str) -> bool:
    return USERNAME_REGEX.match(username_string) is not None



from datetime import datetime, timedelta
from enum import Enum
from typing import List, Optional

from pydantic import Field

from .base import BaseModel


class DurationUnit(Enum):
    seconds = "seconds"
    minutes = "minutes"
    hours = "hours"
    days = "days"
    weeks = "weeks"


class Duration(BaseModel):
    """Representation of time duration in a given set of units.
    Used for specifiying bucket interval sizes, typically in metrics charts."""

    unit: DurationUnit
    value: int


class RunMetric(BaseModel):
    """
    Type of run metric data specified for a pipeline.
    """

    #: total number of runs executed
    run_count: int
    #: total number of failed runs
    failed_run_count: int
    #: total number of succeeded runs
    succeeded_run_count: int
    #: total run compute time (typically of succeeded runs)
    total_compute_ms: int


class RunHardwareMetric(BaseModel):
    resource_type: str
    total_percentage: float
    run_count: int
    average_runtime: int


class HardwareMetric(BaseModel):
    start: datetime
    end: datetime
    account_usage: List[RunHardwareMetric]


class RunMetricsBucket(BaseModel):
    """
    Type of metadata of a specific bucket.

    start: start timestamp of the time range
    end: end timestamp of the time range
    run_count: count of runs part of this bucket
    success_rate: percentage of successful runs as float between 0 and 1
    average_runtime: average runtime of all runs in this bucket
    total_runtime: total runtime of all runs in this bucket
    """

    start: datetime
    end: datetime
    run_count: int
    success_rate: float
    average_runtime: int
    total_runtime: int


class RunMetricsGet(BaseModel):
    """
    Return type of the metrics endpoint on run related requests.

    start: start timestamp of the time range
    end: end timestamp of the time range
    bucket_count: number of buckets time range was divided into
    metrics_buckets: list of metadata about each bucket
    overall_bucket: metadata for the whole time range
    preceding_bucket: metadata for the bucket directly preceding the time range
    """

    start: datetime
    end: datetime
    bucket_count: int
    metrics_buckets: List[RunMetricsBucket]
    overall_bucket: RunMetricsBucket
    preceding_bucket: RunMetricsBucket


class PipelineMetricsGet(RunMetricsGet):
    """Run metrics supplemented with pipeline meta data. Used in particular
    for list/paginated metrics views where it is important for the client to know which
    pipeline is which"""

    pipeline_id: str
    pipeline_name: str


class PipelineMetricsGetSummary(RunMetric):
    """Pipeline run summary metrics supplemented with pipeline meta data.
    Pipeline meta data is useful for list/paginated metrics views so the client
    can keep track of which pipeline is which"""

    pipeline_id: str
    pipeline_name: str


class MetricsQuery(BaseModel):
    """
    Query parameters for generic metrics requests.

    start: start timestamp of the time range
    end: end timestamp of the time range
    """

    start: datetime = Field(
        default_factory=lambda: datetime.utcnow() - timedelta(hours=24)
    )
    end: datetime = Field(default_factory=datetime.utcnow)


class RunMetricsQuery(MetricsQuery):
    """
    Query parameters for run metrics requests.

    start: start timestamp of the time range
    end: end timestamp of the time range
    bucket_count: number of buckets to divide the time range into
    """

    bucket_count: int = 100


class MetricsBucketsIntervalQuery(MetricsQuery):
    """
    Alternate query parameters for run metrics requests.
    Specify a bucket time interval/duration instead of a number of buckets
    """

    interval: Duration


class PipelineComputeGet(BaseModel):
    """
    Pipeline compute metrics for retrieving the number of completed runs
    and total compute time of all the runs on a pipeline
    """

    start: datetime
    end: datetime
    pipeline_id: str
    pipeline_name: str
    completed_run_count: int
    total_compute_ms: int


class TotalComputeGet(RunMetric):
    """Total compute time of all a users' runs"""

    start: datetime
    end: datetime
    #: change rate of run count
    run_count_rate: Optional[float]
    #: change rate of succeeded run count
    succeeded_run_count_rate: Optional[float]
    #: change rate of failed run count
    failed_run_count_rate: Optional[float]
    #: change rate of compute run time
    total_compute_ms_rate: Optional[float]


class PipelineRunMetricsData(BaseModel):
    """
    Overall and per-timestamp run metrics data for a given pipeline.
    Used as list item in return data for run metrics over multiple pipelines
    """

    pipeline_id: str
    pipeline_name: str
    overall_bucket: RunMetric
    metrics_buckets: List[RunMetric]


class PipelinesRunMetricsGet(MetricsQuery):
    """
    Response schema for retrieving run metrics (overall + per-timestamp)
    on a list of pipelines.
    """

    timestamps: List[datetime]
    data: List[PipelineRunMetricsData]



from typing import List, Optional

from pydantic import validator

from .base import AvatarHolder, BaseModel, Patchable
from .token import TokenGet
from .validators import valid_email, valid_password, valid_username


class UserBase(AvatarHolder):
    email: str
    username: str
    firstname: Optional[str]
    lastname: Optional[str]
    company: Optional[str]
    job_title: Optional[str]


class UserGet(UserBase):
    id: str
    oauth_provider: Optional[str]
    verified: Optional[bool]
    subscribed: Optional[bool]
    onboarded: Optional[bool]
    preferences: Optional[dict]


class UserGetDetailed(UserGet):
    tokens: List[TokenGet] = []


class UserGetEnriched(UserGetDetailed):
    base_token: TokenGet


class UserPatch(Patchable, AvatarHolder):
    firstname: Optional[str]
    lastname: Optional[str]
    company: Optional[str]
    job_title: Optional[str]
    subscribed: Optional[bool]
    onboarded: Optional[bool]


class UserUsernamePatch(Patchable):
    username: str

    @validator("username")
    def validate_username(cls, value):
        if not valid_username(value):
            raise ValueError(
                (
                    "must contain between 3-24 characters, only alphanumerics, "
                    "hyphens and underscores."
                )
            )
        return value


class UserEmailPatch(Patchable):
    email: str

    @validator("email")
    def validate_email(cls, value):
        lowered_value = value.lower()
        if not valid_email(lowered_value):
            raise ValueError("doesn't match standard email format")
        return lowered_value


class UserPasswordPatch(Patchable):
    old_password: str
    password: str

    @validator("password")
    def validate_password(cls, value):
        if not valid_password(value):
            raise ValueError(
                (
                    "must contain at least 8 characters, "
                    "one uppercase letter and one number."
                )
            )
        return value


class UserPasswordResetPatch(Patchable):
    password: str

    @validator("password")
    def validate_password(cls, value):
        if not valid_password(value):
            raise ValueError(
                (
                    "must contain at least 8 characters, "
                    "one uppercase letter and one number."
                )
            )
        return value


class UserLogin(BaseModel):
    email: str
    password: str

    @validator("email")
    def validate_email(cls, value):
        lowered_value = value.lower()
        if not valid_email(lowered_value):
            raise ValueError("doesn't match standard email format")
        return lowered_value


class UserOAuthLogin(BaseModel):
    email: str
    oauth_id: str
    oauth_provider: str


class UserCreate(UserBase):
    password: str
    username: Optional[str]
    account_type: Optional[str]

    @validator("email")
    def validate_email(cls, value):
        lowered_value = value.lower()
        if not valid_email(lowered_value):
            raise ValueError("doesn't match standard email format")
        return lowered_value

    @validator("password")
    def validate_password(cls, value):
        if not valid_password(value):
            raise ValueError(
                (
                    "must contain at least 8 characters, "
                    "one uppercase letter and one number."
                )
            )
        return value

    @validator("username")
    def validate_username(cls, value):
        if value is not None and not valid_username(value):
            raise ValueError(
                (
                    "must contain between 3-24 characters, "
                    "only alphanumerics, hyphens and underscores."
                )
            )
        return value
